import itertools
from typing import Dict, List, Sequence, Union, Optional
import os
import copy
import json
import os.path as osp
from pathlib import Path
from collections import OrderedDict

import torch
from prettytable import PrettyTable
import numpy as np

from mmengine.evaluator import BaseMetric
from mmengine.logging import MMLogger
from mmengine import load
from mmengine.registry import METRICS
from mmpretrain.evaluation.metrics import SingleLabelMetric
from visaionlibrary.datasets.utils import encode_mask_to_str



def calc_iou(bbox1, bbox2):
    if not isinstance(bbox1, np.ndarray):
        bbox1 = np.array(bbox1)
    if not isinstance(bbox2, np.ndarray):
        bbox2 = np.array(bbox2)
    xmin1, ymin1, xmax1, ymax1, = np.split(bbox1, 4, axis=-1)
    xmin2, ymin2, xmax2, ymax2, = np.split(bbox2, 4, axis=-1)

    area1 = (xmax1 - xmin1) * (ymax1 - ymin1)
    area2 = (xmax2 - xmin2) * (ymax2 - ymin2)

    ymin = np.maximum(ymin1, np.squeeze(ymin2, axis=-1))
    xmin = np.maximum(xmin1, np.squeeze(xmin2, axis=-1))
    ymax = np.minimum(ymax1, np.squeeze(ymax2, axis=-1))
    xmax = np.minimum(xmax1, np.squeeze(xmax2, axis=-1))

    h = np.maximum(ymax - ymin, 0)
    w = np.maximum(xmax - xmin, 0)
    intersect = h * w

    union = area1 + np.squeeze(area2, axis=-1) - intersect + 1e-6
    return intersect / union


@METRICS.register_module()
class PrecisionAndRecallsMetric(BaseMetric):
    """ Output Overkill and Miss detections.
    detail ref mmdet.evaluation.CocoMetric
    compute F1 score and
    """


    METRIC_NAMES = ['detect_correct', 'wrong_detect', 'overkill', 'escape', 'detect_correct_rate',
               'escape_rate', 'overkill_rate', 'best_confidence', 'F1_score']
    def __init__(self,
                 iou_threshold=0.3,
                 min_area=64,
                 metric_items: List = None,
                 collect_device: str = 'cpu',
                 prefix: Optional[str] = None,
                 ) -> None:
        """
           Args:
               confidence_threshold:
               iou_threshold:
               metric_items: ['detect_correct'  # tp
                               'wrong_detect' # mis
                               'overkill'  # fp
                               'escape # fn
                               'detect_correct_rate': 4,  # tp / (tp+mis+fn)
                               'escape_rate': 5,  # fn / (tp+mis+fn)
                               'overkill_rate': 6,  # fp / (tp+fp)
                               'best_confidence': 7,
                               'F1_score': 8
           """
        super().__init__(collect_device=collect_device, prefix=prefix)
        assert iou_threshold >= 0
        self.iou_threshold = iou_threshold
        self.results_extend = {}
        self.min_area = min_area
        self.num_class = None

        if not metric_items:
            metric_items = ['detect_correct_rate', 'escape_rate', 'overkill_rate', 'best_confidence', 'F1_score']
        for metric_item in metric_items:
            assert metric_item in self.METRIC_NAMES, f'{metric_item} must in list {METRICS}'
        self.metric_items = metric_items

    def process(self, data_batch: dict, data_samples: Sequence[dict]) -> None:
        """Process one batch of data samples and predictions. The processed
        results should be stored in ``self.results``, which will be used to
        compute the metrics when all batches have been processed.

        Args:
            data_batch (dict): A batch of data from the dataloader.
            data_samples (Sequence[dict]): A batch of data samples that
                contain annotations and predictions.
        """
        if self.num_class is None:
            self.num_class = len(self.dataset_meta['classes'])

        # for循环一个batch中的所有图
        for data_sample in data_samples:

            # 预测的所有bbox
            pred_instances = data_sample.get("pred_instances", torch.Tensor([]))
            pred_bboxes_ = pred_instances.get('bboxes', torch.Tensor([])).detach().cpu().numpy()    # [N', 4], N'表示预测的bbox数量
            pred_scores_ = pred_instances['scores'].detach().cpu().numpy()                          # [N']
            pred_labels_ = pred_instances['labels'].detach().cpu().numpy()                          # [N']

            # 真实gt
            gt_bboxes, gt_labels = [], []
            for isinstance in data_sample['instances']:
                gt_bboxes.append(isinstance['bbox'])
                gt_labels.append(isinstance['bbox_label'])

            gt_bboxes = np.array(gt_bboxes, dtype=pred_bboxes_.dtype)   # [N, 4]
            gt_labels = np.array(gt_labels, dtype=pred_labels_.dtype)   # [N]

            # 计算iou
            if len(gt_bboxes) == 0:
                # tp list append 0
                # fp list append len(pred_bbox)
                gt_bboxes = np.array([[0, 0, 0, 0]])
                gt_labels = [None]
            # filter results by confidence threshold [ldf]
            # for each confidence
            for conf_th in range(1, 100):
                score_filter = pred_scores_ >= conf_th * 0.01
                pred_bboxes = pred_bboxes_[score_filter]    # [N'', 4], 过滤掉置信度小于conf_th的bbox
                pred_labels = pred_labels_[score_filter]    # [N'']
                pred_scores = pred_scores_[score_filter]    # [N'']

                if len(pred_bboxes) == 0:
                    pred_bboxes = np.array([[0, 0, 0, 0]])
                    pred_labels = [None]
                    pred_scores = [0]
                ious_argmax = self.match_gt_pred(gt_bboxes, pred_bboxes, pred_scores, gt_labels, pred_labels)

                pred_labels_extend = []
                gt_labels_extend = []
                # label self.num_class is background, -1 is ignore label usually
                for gt_index, gt_i in enumerate(gt_labels): # 遍历gt的label
                    if gt_index in ious_argmax.keys():  # 如果gt index在  gt和pred的匹配中, 取出预测的pred label, 情况1, gt和pred匹配上了, 但不知匹配是否正确
                        pred_labels_extend.append(pred_labels_[ious_argmax[gt_index]])
                        gt_labels_extend.append(gt_i)
                    elif gt_i is not None:  # 如果gt_id 不在gt和pred的匹配中, 将pred的label 置为 一个不存在的类别, 情况2, gt和pred不匹配, 漏检, 有gt没有pred
                        pred_labels_extend.append(self.num_class)
                        gt_labels_extend.append(gt_i)
                for pred_index, pred_i in enumerate(pred_labels): # 遍历pred的label
                    if pred_index not in ious_argmax.values() and pred_i is not None:  # 如果pred index 不在gt和pred的匹配中, 将gt的label 置为 一个不存在的类别, 情况3, pred和gt不匹配, 过检, 有pred没有gt
                        pred_labels_extend.append(pred_i)
                        gt_labels_extend.append(self.num_class)
                if conf_th in self.results_extend.keys():
                    self.results_extend[conf_th].append(
                        {
                            'pred_labels_extend': pred_labels_extend,
                            'gt_labels_extend': gt_labels_extend
                        }
                    )
                else:
                    self.results_extend[conf_th] = [
                        {
                            'pred_labels_extend': pred_labels_extend,
                            'gt_labels_extend': gt_labels_extend
                        }
                    ]

    def match_gt_pred(self, gb, pb, ps, gl, pl) -> dict:    # gt_bboxes, pred_bboxes, pred_scores, gt_labels, pred_labels
        ious = calc_iou(gb, pb)  # gb是gt bbox->[30, 4], pb是pred bbox->[67, 4], eg. (30, 67),表示gt对应pred的iou值
        match_dict = {}
        # filter with iou_threshold
        ious = ious * (ious > self.iou_threshold)  # (n, m), 过滤掉iou小于阈值的bbox
        # filter with label
        ious = ious * (np.tile(np.expand_dims(gl, 1), (1, ious.shape[1])) == np.tile(pl, (ious.shape[0], 1))) # 过滤掉label不匹配的bbox, iou shape->(30, 67)
        index = np.where(ious > 0)  # 返回gt和pred的索引, index[0]是gt的索引,也就是行, index[1]是pred的索引,也就是列, 过滤后index[0].shape->34, index[1].shape->34
        if not len(index[0]):
            return match_dict
        matches = np.array([ious[index], ps[index[1]], index[0], index[1]]).T   # shape->(34, 4), 34表示匹配的bbox对数, 4表示iou, pred score, gt index, pred index
        # order by pred score and iou score
        matches = matches[np.lexsort((-matches[:, 0], -matches[:, 1]))]  # 先按iou降序, 再按pred score降序

        for i, match in enumerate(matches):
            if match[2] not in match_dict.keys():   # key是gt的index, value是pred的index
                match_dict[int(match[2])] = int(match[3])

        return match_dict  # {gt_index: pred_index}, 表示gt的index和pred的index的对应关系

    def compute_metrics(self, results: list) -> Dict[str, float]:
        """Compute the metrics from processed results.

        Args:
            results (list): The processed results of each batch.

        Returns:
            Dict[str, float]: The computed metrics. The keys are the names of
            the metrics, and the values are corresponding results.
        """

        logger: MMLogger = MMLogger.get_current_instance()
        eval_results = OrderedDict()

        metric_names = {
            'detect_correct': 0,  # tp
            'wrong_detect': 1,  # mis
            'overkill': 2,  # fp
            'escape': 3,  # fn
            'detect_correct_rate': 4,  # tp / (tp+mis+fn)
            'escape_rate': 5,  # fn / (tp+mis+fn)
            'overkill_rate': 6,  # fp / (tp+fp)
            'best_confidence': 7,
            'F1_score': 8
        }

        # compute f1 score
        f1_scores = []
        slm = SingleLabelMetric(average='micro')
        labels = {}
        for conf in range(1, 100):
            # compute most f1
            if conf not in labels.keys():
                labels[conf] = {
                    'y_pred': [],
                    'y_true': []
                }
            for item in self.results_extend[conf]:
                labels[conf]['y_pred'].extend(item['pred_labels_extend'])
                labels[conf]['y_true'].extend(item['gt_labels_extend'])
            precision, recall, f1_score_, support = slm.calculate(np.array(labels[conf]['y_pred']),
                                                                  np.array(labels[conf]['y_true']),
                                                                  average='micro', num_classes=self.num_class+1)
            f1_scores.append(f1_score_)
        f1_score = float(max(f1_scores))
        conf_index = f1_scores.index(f1_score) + 1
        y_pred, y_true = labels[conf_index]['y_pred'], labels[conf_index]['y_true']
        tp = sum([1 if y_pred[i] == y_true[i] else 0 for i in range(len(y_pred))])
        fp = sum([1 if y_pred[i] != y_true[i] and y_true[i] == self.num_class else 0 for i in range(len(y_pred))])
        fn = sum([1 if y_pred[i] != y_true[i] and y_pred[i] == self.num_class else 0 for i in range(len(y_pred))])
        mis = sum([1 if y_pred[i] != y_true[i] and y_pred[i] != self.num_class and y_true[i] != self.num_class else 0
                   for i in range(len(y_pred))])

        result_list = [tp, mis, fp, fn, tp / (tp+mis+fn+1e-6), fn / (tp+mis+fn+1e-6),
                       fp / (tp+fp+1e-6), conf_index*0.01, f1_score]
        headers = ['检测正确', '检错类别', '过检', '漏检',
                   '检出率', '漏检率', '过检率', '最佳置信度', 'f1_score']
        table_data = PrettyTable()

        # filter results
        for i, name in enumerate(metric_names):
            if name in self.metric_items:
                eval_results[name] = result_list[i]
                if i in [4, 5, 6, 7]:
                    table_data.add_column(headers[i], [f'{result_list[i]*100: .2f}%'])
                elif i == 8:
                    table_data.add_column(headers[i], [f'{result_list[i]: .2f}%'])
                else:
                    table_data.add_column(headers[i], [f'{result_list[i]: .2f}'])

        logger.info('\n' + table_data.get_string())

        self.results_extend.clear()
        return eval_results


@METRICS.register_module()
class PrecisionAndRecallsMetricforVisaion(PrecisionAndRecallsMetric):
    def __init__(self,
                 confidence_threshold: float = 0.05,
                 iou_threshold: float = 0.5,
                 output_dir: Optional[str] = None,
                 **kwargs) -> None:

        super().__init__(iou_threshold=iou_threshold)
        self.confidence_threshold = confidence_threshold

        # 设置输出目录
        if os.environ.get("VISAION_WORK_DIR", None):
            # 设置输出目录
            self.output_dir = os.environ.get("VISAION_WORK_DIR", None)
        else:
            self.output_dir = output_dir
        assert self.output_dir is not None
        os.makedirs(self.output_dir, exist_ok=True)

        # pred_origin
        self.output_dir_pred_origin = osp.join(self.output_dir, "pred_origin")
        os.makedirs(self.output_dir_pred_origin, exist_ok=True)

        # instance_paring
        self.output_dir_instance_paring = osp.join(self.output_dir, "instance_paring")
        os.makedirs(self.output_dir_instance_paring, exist_ok=True)

    def xyxy2xywh(self, bbox: np.ndarray) -> list:
        """Convert ``xyxy`` style bounding boxes to ``xywh`` style for COCO
        evaluation.

        Args:
            bbox (numpy.ndarray): The bounding boxes, shape (4, ), in
                ``xyxy`` order.

        Returns:
            list[float]: The converted bounding boxes, in ``xywh`` order.
        """

        _bbox: List = bbox.tolist()
        return [
            _bbox[0],
            _bbox[1],
            _bbox[2] - _bbox[0],
            _bbox[3] - _bbox[1],
        ]

    def process(self, data_batch: dict, data_samples: Sequence[dict]) -> None:
        """Process one batch of data samples and predictions. The processed
        results should be stored in ``self.results``, which will be used to
        compute the metrics when all batches have been processed.
        Output:
        - instance_paring
            |- 1.json
            |- 2.json
            ...
        Inside 1.json:
        [
            [{'gtDefectId': None, 'predDefectId': 0, 'count': 3},
            {'gtDefectId': 0, 'predDefectId': None, 'count': 1},
            ],  # for confidence 0.01
            [...]  # for confidence 0.02, ...
        ]

        Args:
            data_batch (dict): A batch of data from the dataloader.
            data_samples (Sequence[dict]): A batch of data samples that
                contain annotations and predictions.
        """
        super().process(data_batch=data_batch, data_samples=data_samples)

        # for循环一个batch中的所有图
        for data_sample in data_samples:
            sample_id = Path(data_sample['img_path']).stem    # 使用图片名作为sample_id
            # sample_id = data_sample['img_id']  # 使用图片id作为sample_id


            # 预测的所有bbox
            pred_instances = data_sample.get("pred_instances", torch.Tensor([]))
            pred_bboxes = pred_instances.get('bboxes', torch.Tensor([]))
            pred_bboxes = pred_bboxes.detach().cpu().numpy()
            pred_scores = pred_instances['scores'].detach().cpu().numpy()
            pred_labels = pred_instances['labels'].detach().cpu().numpy()
            if "masks" in pred_instances.keys():
                pred_masks = pred_instances['masks'].detach().cpu().numpy()

            # filter results by confidence threshold [ldf]
            score_filter = pred_scores >= self.confidence_threshold
            pred_bboxes = pred_bboxes[score_filter]
            pred_labels = pred_labels[score_filter]
            pred_scores = pred_scores[score_filter]
            if "masks" in pred_instances.keys():
                pred_masks = pred_masks[score_filter]

            # 存储预测结果(原始)
            pred_dict_origin = dict()
            pred_dict_origin['bboxes'] = pred_bboxes
            pred_dict_origin['labels'] = pred_labels
            pred_dict_origin['scores'] = pred_scores
            if "masks" in pred_instances.keys():
                pred_dict_origin['masks'] = pred_masks
            save_dir = osp.join(self.output_dir_pred_origin, str(sample_id) + ".npy")
            np.save(save_dir, pred_dict_origin)

            # 存储预测结果(显示), 用于平台可视化
            pred_dict_show_list = list()
            for i, (bbox, label, score) in enumerate(zip(pred_bboxes, pred_labels, pred_scores)):
                pred_dict_show = dict()
                pred_dict_show['boundingBox'] = self.xyxy2xywh(bbox)
                pred_dict_show['score'] = score.item()
                pred_dict_show['name'] = self.dataset_meta['classes'][label.item()]
                if "masks" in pred_instances.keys():
                    pred_dict_show['mask'] = encode_mask_to_str(pred_masks[i]*1)
                pred_dict_show_list.append(pred_dict_show)
            save_dir = osp.join(self.output_dir_pred_origin, str(sample_id) + ".json")
            self.save_json(pred_dict_show_list, save_dir, same_file_mode='append')

            # 真实gt
            gt_bboxes, gt_labels = [], []
            for isinstance in data_sample['instances']:
                gt_bboxes.append(isinstance['bbox'])
                gt_labels.append(isinstance['bbox_label'])
            gt_bboxes = np.array(gt_bboxes, dtype=pred_bboxes.dtype)
            gt_labels = np.array(gt_labels, dtype=pred_labels.dtype)

            # 计算iou
            if len(gt_bboxes) == 0:
                gt_bboxes = np.array([[0, 0, 0, 0]])
                gt_labels = [None]
            if len(pred_bboxes) == 0:
                pred_bboxes = np.array([[0, 0, 0, 0]])
                pred_labels = [None]
                pred_scores = [0]
            ious_argmax = self.match_gt_pred(gt_bboxes, pred_bboxes, pred_scores, gt_labels, pred_labels)

            # mutli-class
            instance_pairs = []  # gt和pred对
            for gt_index, gt_i in enumerate(gt_labels):  # 这张图上所有gt循环
                if gt_index in ious_argmax.keys() and gt_i == pred_labels[ious_argmax[gt_index]].item():  # gt and pred
                    temp_dict = dict()
                    temp_dict['score'] = pred_scores[ious_argmax[gt_index]]
                    temp_dict['DefectId'] = gt_i
                    temp_dict['gt_nums'] = 1
                    temp_dict['pred_nums'] = 1
                    temp_dict['tp_nums'] = 1
                elif gt_index in ious_argmax.keys() and gt_i != pred_labels[ious_argmax[gt_index]].item():
                    # count 2 instance
                    temp_dict = dict()
                    temp_dict['score'] = pred_scores[ious_argmax[gt_index]]
                    temp_dict['DefectId'] = gt_i
                    temp_dict['gt_nums'] = 1
                    temp_dict['pred_nums'] = 0
                    temp_dict['tp_nums'] = 0
                    instance_pairs.append(temp_dict)
                    temp_dict = dict()
                    temp_dict['score'] = pred_scores[ious_argmax[gt_index]]
                    temp_dict['DefectId'] = pred_labels[ious_argmax[gt_index]].item()
                    temp_dict['gt_nums'] = 0
                    temp_dict['pred_nums'] = 1
                    temp_dict['tp_nums'] = 0
                else:  # gt not pred
                    temp_dict = dict()
                    temp_dict['score'] = None
                    temp_dict['DefectId'] = gt_i
                    temp_dict['gt_nums'] = 1 if gt_i is not None else 0
                    temp_dict['pred_nums'] = 0
                    temp_dict['tp_nums'] = 0
                instance_pairs.append(temp_dict)
            for pred_index, pred_i in enumerate(pred_labels):  # 这张图上所有的pred循环
                if pred_index not in ious_argmax.values():  # pred not gt
                    temp_dict = dict()
                    temp_dict['score'] = pred_scores[pred_index]
                    temp_dict['DefectId'] = pred_i.item() if pred_i is not None else pred_i
                    temp_dict['gt_nums'] = 0
                    temp_dict['pred_nums'] = 1 if pred_i is not None else 0
                    temp_dict['tp_nums'] = 0
                    instance_pairs.append(temp_dict)

            # single-class
            for gt_index, gt_i in enumerate(gt_labels):  # 这张图上所有gt循环
                if gt_index in ious_argmax.keys():  # gt and pred
                    temp_dict = dict()
                    temp_dict['score'] = pred_scores[ious_argmax[gt_index]]
                    temp_dict['DefectId'] = -1 if gt_i is not None else gt_i
                    temp_dict['gt_nums'] = 1 if gt_i is not None else 0
                    temp_dict['pred_nums'] = 1
                    temp_dict['tp_nums'] = 1 if gt_i is not None else 0
                else:  # gt not pred
                    temp_dict = dict()
                    temp_dict['score'] = None
                    temp_dict['DefectId'] = -1 if gt_i is not None else gt_i
                    temp_dict['gt_nums'] = 1 if gt_i is not None else 0
                    temp_dict['pred_nums'] = 0
                    temp_dict['tp_nums'] = 0
                instance_pairs.append(temp_dict)
            for pred_index, pred_i in enumerate(pred_labels):  # 这张图上所有的pred循环
                if pred_index not in ious_argmax.values():  # pred not gt
                    temp_dict = dict()
                    temp_dict['score'] = pred_scores[pred_index]
                    temp_dict['DefectId'] = -1
                    temp_dict['gt_nums'] = 0
                    temp_dict['pred_nums'] = 1
                    temp_dict['tp_nums'] = 0
                    instance_pairs.append(temp_dict)

            # 统计 gt和pred预测对 在不同置信度下的个数
            instance_paring = []
            for i in range(0, 100):  # 0.0~0.99
                conf_i = i * 0.01
                instance_paring_conf_i = dict()     # 置信度conf_i下的预测对
                instance_pairs_tmp = copy.deepcopy(instance_pairs)
                for pair_ in instance_pairs_tmp:
                    score = pair_.pop('score', 0)
                    if score is None:
                        score = 0
                    if score < conf_i:
                        pair_['pred_nums'] = 0
                        pair_['tp_nums'] = 0
                    if pair_['DefectId'] is None:
                        continue
                    if pair_['DefectId'] in instance_paring_conf_i.keys():
                        instance_paring_conf_i[pair_['DefectId']]['gt_nums'] += pair_['gt_nums']
                        instance_paring_conf_i[pair_['DefectId']]['pred_nums'] += pair_['pred_nums']
                        instance_paring_conf_i[pair_['DefectId']]['tp_nums'] += pair_['tp_nums']
                    else:
                        instance_paring_conf_i[pair_['DefectId']] = dict()
                        instance_paring_conf_i[pair_['DefectId']]['gt_nums'] = pair_['gt_nums']
                        instance_paring_conf_i[pair_['DefectId']]['pred_nums'] = pair_['pred_nums']
                        instance_paring_conf_i[pair_['DefectId']]['tp_nums'] = pair_['tp_nums']

                instance_paring_conf_i_result = []
                for _id in instance_paring_conf_i.keys():
                    tmp_dict_ = {
                        'DefectId': int(_id),
                        'gt_nums': instance_paring_conf_i[_id]['gt_nums'],
                        'pred_nums': instance_paring_conf_i[_id]['pred_nums'],
                        'tp_nums': instance_paring_conf_i[_id]['tp_nums']
                    }
                    instance_paring_conf_i_result.append(tmp_dict_)
                instance_paring.append(instance_paring_conf_i_result)
            # actrually used
            save_dir = osp.join(self.output_dir_instance_paring, str(sample_id) + ".json")
            self.save_json(instance_paring, save_dir, same_file_mode='add')

    @staticmethod
    def save_json(json_dict: Union[Dict, List] = None, save_path: str = None, same_file_mode: str = None) -> None:
        """
        保存dict为json文件
        :param json_dict:
        :param save_path:   保存路径
        :same_file_mode: default is update
        :return:
        """

        # 如果文件里面已经有东西了，拿出来追加
        if osp.exists(save_path):
            ori_info = load(save_path)  # [{}, {}]
            # for each ins add
            if same_file_mode == 'add':
                for i, ins in enumerate(json_dict):
                    if isinstance(ins, List):
                        json_dict[i].extend(ori_info[i])
            elif same_file_mode == 'append':
                json_dict.extend(ori_info)
        with open(save_path, 'w', encoding="utf-8") as fp:
            json.dump(json_dict, fp, ensure_ascii=False, indent=4)



import warnings
from abc import ABCMeta, abstractmethod
from typing import List, Tuple

import torch
import torch.nn as nn
from mmengine.model import BaseModule
from torch import Tensor

from mmseg.structures import build_pixel_sampler
from mmseg.utils import ConfigType, SampleList
from mmseg.models.builder import build_loss
from mmseg.models.losses import accuracy
from mmseg.models.utils import resize


class BaseDecodeHead(BaseModule, metaclass=ABCMeta):
    """Base class for BaseDecodeHead.

    1. The ``init_weights`` method is used to initialize decode_head's
    model parameters. After segmentor initialization, ``init_weights``
    is triggered when ``segmentor.init_weights()`` is called externally.

    2. The ``loss`` method is used to calculate the loss of decode_head,
    which includes two steps: (1) the decode_head model performs forward
    propagation to obtain the feature maps (2) The ``loss_by_feat`` method
    is called based on the feature maps to calculate the loss.

    .. code:: text

    loss(): forward() -> loss_by_feat()

    3. The ``predict`` method is used to predict segmentation results,
    which includes two steps: (1) the decode_head model performs forward
    propagation to obtain the feature maps (2) The ``predict_by_feat`` method
    is called based on the feature maps to predict segmentation results
    including post-processing.

    .. code:: text

    predict(): forward() -> predict_by_feat()

    Args:
        in_channels (int|Sequence[int]): Input channels.
        channels (int): Channels after modules, before conv_seg.
        num_classes (int): Number of classes.
        out_channels (int): Output channels of conv_seg.
        threshold (float): Threshold for binary segmentation in the case of
            `num_classes==1`. Default: None.
        dropout_ratio (float): Ratio of dropout layer. Default: 0.1.
        conv_cfg (dict|None): Config of conv layers. Default: None.
        norm_cfg (dict|None): Config of norm layers. Default: None.
        act_cfg (dict): Config of activation layers.
            Default: dict(type='ReLU')
        in_index (int|Sequence[int]): Input feature index. Default: -1
        input_transform (str|None): Transformation type of input features.
            Options: 'resize_concat', 'multiple_select', None.
            'resize_concat': Multiple feature maps will be resize to the
                same size as first one and than concat together.
                Usually used in FCN head of HRNet.
            'multiple_select': Multiple feature maps will be bundle into
                a list and passed into decode head.
            None: Only one select feature map is allowed.
            Default: None.
        loss_decode (dict | Sequence[dict]): Config of decode loss.
            The `loss_name` is property of corresponding loss function which
            could be shown in training log. If you want this loss
            item to be included into the backward graph, `loss_` must be the
            prefix of the name. Defaults to 'loss_ce'.
             e.g. dict(type='CrossEntropyLoss'),
             [dict(type='CrossEntropyLoss', loss_name='loss_ce'),
              dict(type='DiceLoss', loss_name='loss_dice')]
            Default: dict(type='CrossEntropyLoss').
        ignore_index (int | None): The label index to be ignored. When using
            masked BCE loss, ignore_index should be set to None. Default: 255.
        sampler (dict|None): The config of segmentation map sampler.
            Default: None.
        align_corners (bool): align_corners argument of F.interpolate.
            Default: False.
        init_cfg (dict or list[dict], optional): Initialization config dict.
    """

    def __init__(self,
                 in_channels,
                 channels,
                 *,
                 num_classes,
                 out_channels=None,
                 threshold=None,
                 dropout_ratio=0.1,
                 conv_cfg=None,
                 norm_cfg=None,
                 act_cfg=dict(type='ReLU'),
                 in_index=-1,
                 input_transform=None,
                 sigmoid_before_loss=False,
                 loss_decode=dict(
                     type='CrossEntropyLoss',
                     use_sigmoid=False,
                     loss_weight=1.0),
                 ignore_index=255,
                 sampler=None,
                 align_corners=False,
                 init_cfg=dict(
                     type='Normal', std=0.01, override=dict(name='conv_seg'))):
        super().__init__(init_cfg)
        self._init_inputs(in_channels, in_index, input_transform)
        self.channels = channels
        self.dropout_ratio = dropout_ratio
        self.conv_cfg = conv_cfg
        self.norm_cfg = norm_cfg
        self.act_cfg = act_cfg
        self.in_index = in_index

        self.ignore_index = ignore_index
        self.align_corners = align_corners

        assert num_classes >= 2, 'at least 2 classes have to be set (background class is added by default)'
        if out_channels is None:
            if num_classes == 2:
                warnings.warn('For binary segmentation, we suggest using'
                              '`out_channels = 1` to define the output'
                              'channels of segmentor, and use `threshold`'
                              'to convert `seg_logits` into a prediction'
                              'applying a threshold')
            out_channels = num_classes

        if out_channels != num_classes and out_channels != 1:
            raise ValueError(
                'out_channels should be equal to num_classes,'
                'except binary segmentation set out_channels == 1 and'
                f'num_classes == 2, but got out_channels={out_channels}'
                f'and num_classes={num_classes}')

        if out_channels == 1 and threshold is None:
            threshold = 0.3
            warnings.warn('threshold is not defined for binary, and defaults'
                          'to 0.3')
        self.num_classes = num_classes
        self.out_channels = out_channels
        self.threshold = threshold

        if isinstance(loss_decode, dict):
            self.loss_decode = build_loss(loss_decode)
        elif isinstance(loss_decode, (list, tuple)):
            self.loss_decode = nn.ModuleList()
            for loss in loss_decode:
                self.loss_decode.append(build_loss(loss))
        else:
            raise TypeError(f'loss_decode must be a dict or sequence of dict,\
                but got {type(loss_decode)}')

        # this variable controls the sigmoid activation before loss
        # it affects prediction and inference
        # it should be compatible with the loss
        self.sigmoid_before_loss = sigmoid_before_loss

        if sampler is not None:
            self.sampler = build_pixel_sampler(sampler, context=self)
        else:
            self.sampler = None

        self.conv_seg = nn.Conv2d(channels, self.out_channels, kernel_size=1)
        if dropout_ratio > 0:
            self.dropout = nn.Dropout2d(dropout_ratio)
        else:
            self.dropout = None

    def extra_repr(self):
        """Extra repr."""
        s = f'input_transform={self.input_transform}, ' \
            f'ignore_index={self.ignore_index}, ' \
            f'align_corners={self.align_corners}'
        return s

    def _init_inputs(self, in_channels, in_index, input_transform):
        """Check and initialize input transforms.

        The in_channels, in_index and input_transform must match.
        Specifically, when input_transform is None, only single feature map
        will be selected. So in_channels and in_index must be of type int.
        When input_transform

        Args:
            in_channels (int|Sequence[int]): Input channels.
            in_index (int|Sequence[int]): Input feature index.
            input_transform (str|None): Transformation type of input features.
                Options: 'resize_concat', 'multiple_select', None.
                'resize_concat': Multiple feature maps will be resize to the
                    same size as first one and than concat together.
                    Usually used in FCN head of HRNet.
                'multiple_select': Multiple feature maps will be bundle into
                    a list and passed into decode head.
                None: Only one select feature map is allowed.
        """

        if input_transform is not None:
            assert input_transform in ['resize_concat', 'multiple_select']
        self.input_transform = input_transform
        self.in_index = in_index
        if input_transform is not None:
            assert isinstance(in_channels, (list, tuple))
            assert isinstance(in_index, (list, tuple))
            assert len(in_channels) == len(in_index)
            if input_transform == 'resize_concat':
                self.in_channels = sum(in_channels)
            else:
                self.in_channels = in_channels
        else:
            assert isinstance(in_channels, int)
            assert isinstance(in_index, int)
            self.in_channels = in_channels

    def _transform_inputs(self, inputs):
        """Transform inputs for decoder.

        Args:
            inputs (list[Tensor]): List of multi-level img features.

        Returns:
            Tensor: The transformed inputs
        """

        if self.input_transform == 'resize_concat':
            inputs = [inputs[i] for i in self.in_index]
            upsampled_inputs = [
                resize(
                    input=x,
                    size=inputs[0].shape[2:],
                    mode='bilinear',
                    align_corners=self.align_corners) for x in inputs
            ]
            inputs = torch.cat(upsampled_inputs, dim=1)
        elif self.input_transform == 'multiple_select':
            inputs = [inputs[i] for i in self.in_index]
        else:
            inputs = inputs[self.in_index]

        return inputs

    @abstractmethod
    def forward(self, inputs):
        """Placeholder of forward function."""
        pass

    def cls_seg(self, feat):
        """Classify each pixel."""
        if self.dropout is not None:
            feat = self.dropout(feat)
        output = self.conv_seg(feat)
        return output

    def loss(self, inputs: Tuple[Tensor], batch_data_samples: SampleList,
             train_cfg: ConfigType) -> dict:
        """Forward function for training.

        Args:
            inputs (Tuple[Tensor]): List of multi-level img features.
            batch_data_samples (list[:obj:`SegDataSample`]): The seg
                data samples. It usually includes information such
                as `img_metas` or `gt_semantic_seg`.
            train_cfg (dict): The training config.

        Returns:
            dict[str, Tensor]: a dictionary of loss components
        """
        seg_logits = self.forward(inputs)
        losses = self.loss_by_feat(seg_logits, batch_data_samples)
        return losses

    def predict(self, inputs: Tuple[Tensor], batch_img_metas: List[dict],
                test_cfg: ConfigType) -> Tensor:
        """Forward function for prediction.

        Args:
            inputs (Tuple[Tensor]): List of multi-level img features.
            batch_img_metas (dict): List Image info where each dict may also
                contain: 'img_shape', 'scale_factor', 'flip', 'img_path',
                'ori_shape', and 'pad_shape'.
                For details on the values of these keys see
                `mmseg/datasets/pipelines/formatting.py:PackSegInputs`.
            test_cfg (dict): The testing config.

        Returns:
            Tensor: Outputs segmentation logits map.
        """
        seg_logits = self.forward(inputs)

        return self.predict_by_feat(seg_logits, batch_img_metas)

    def _stack_batch_gt(self, batch_data_samples: SampleList) -> Tensor:
        gt_semantic_segs = [
            data_sample.gt_sem_seg.data for data_sample in batch_data_samples
        ]
        return torch.stack(gt_semantic_segs, dim=0)

    def loss_by_feat(self, seg_logits: Tensor,
                     batch_data_samples: SampleList) -> dict:
        """Compute segmentation loss.

        Args:
            seg_logits (Tensor): The output from decode head forward function.
            batch_data_samples (List[:obj:`SegDataSample`]): The seg
                data samples. It usually includes information such
                as `metainfo` and `gt_sem_seg`.

        Returns:
            dict[str, Tensor]: a dictionary of loss components
        """

        seg_label = self._stack_batch_gt(batch_data_samples)
        loss = dict()
        seg_logits = resize(
            input=seg_logits,
            size=seg_label.shape[2:],
            mode='bilinear',
            align_corners=self.align_corners)
        if self.sampler is not None:
            seg_weight = self.sampler.sample(seg_logits, seg_label)
        else:
            seg_weight = None
        seg_label = seg_label.squeeze(1)

        if not isinstance(self.loss_decode, nn.ModuleList):
            losses_decode = [self.loss_decode]
        else:
            losses_decode = self.loss_decode
        for loss_decode in losses_decode:
            if loss_decode.loss_name not in loss:
                loss[loss_decode.loss_name] = loss_decode(
                    seg_logits,
                    seg_label,
                    weight=seg_weight,
                    ignore_index=self.ignore_index)
            else:
                loss[loss_decode.loss_name] += loss_decode(
                    seg_logits,
                    seg_label,
                    weight=seg_weight,
                    ignore_index=self.ignore_index)

        loss['acc_seg'] = accuracy(
            seg_logits, seg_label, ignore_index=self.ignore_index)
        return loss

    def predict_by_feat(self, seg_logits: Tensor,
                        batch_img_metas: List[dict]) -> Tensor:
        """Transform a batch of output seg_logits to the input shape.

        Args:
            seg_logits (Tensor): The output from decode head forward function.
            batch_img_metas (list[dict]): Meta information of each image, e.g.,
                image size, scaling factor, etc.

        Returns:
            Tensor: Outputs segmentation logits map.
        """
        if 'pad_shape' in batch_img_metas[0]:
            seg_logits = resize(
                input=seg_logits,
                size=batch_img_metas[0]['pad_shape'],
                mode='bilinear',
                align_corners=self.align_corners)
        else:
            seg_logits = resize(
                input=seg_logits,
                size=batch_img_metas[0]['img_shape'],
                mode='bilinear',
                align_corners=self.align_corners)
        return seg_logits



"""Model validation metrics."""

import math

import numpy as np
import torch


def box_iou(box1, box2, eps=1e-7):
    """
    Calculate intersection-over-union (IoU) of boxes. Both sets of boxes are expected to be in (x1, y1, x2, y2) format.
    Based on https://github.com/pytorch/vision/blob/master/torchvision/ops/boxes.py.

    Args:
        box1 (torch.Tensor): A tensor of shape (N, 4) representing N bounding boxes.
        box2 (torch.Tensor): A tensor of shape (M, 4) representing M bounding boxes.
        eps (float, optional): A small value to avoid division by zero. Defaults to 1e-7.

    Returns:
        (torch.Tensor): An NxM tensor containing the pairwise IoU values for every element in box1 and box2.
    """
    # NOTE: Need .float() to get accurate iou values
    # inter(N,M) = (rb(N,M,2) - lt(N,M,2)).clamp(0).prod(2)
    (a1, a2), (b1, b2) = box1.float().unsqueeze(1).chunk(2, 2), box2.float().unsqueeze(0).chunk(2, 2)
    inter = (torch.min(a2, b2) - torch.max(a1, b1)).clamp_(0).prod(2)

    # IoU = inter / (area1 + area2 - inter)
    return inter / ((a2 - a1).prod(2) + (b2 - b1).prod(2) - inter + eps)


def bbox_iou(box1, box2, xywh=True, GIoU=False, DIoU=False, CIoU=False, eps=1e-7):
    """
    Calculate Intersection over Union (IoU) of box1(1, 4) to box2(n, 4).

    Args:
        box1 (torch.Tensor): A tensor representing a single bounding box with shape (1, 4).
        box2 (torch.Tensor): A tensor representing n bounding boxes with shape (n, 4).
        xywh (bool, optional): If True, input boxes are in (x, y, w, h) format. If False, input boxes are in
                               (x1, y1, x2, y2) format. Defaults to True.
        GIoU (bool, optional): If True, calculate Generalized IoU. Defaults to False.
        DIoU (bool, optional): If True, calculate Distance IoU. Defaults to False.
        CIoU (bool, optional): If True, calculate Complete IoU. Defaults to False.
        eps (float, optional): A small value to avoid division by zero. Defaults to 1e-7.

    Returns:
        (torch.Tensor): IoU, GIoU, DIoU, or CIoU values depending on the specified flags.
    """
    # Get the coordinates of bounding boxes
    if xywh:  # transform from xywh to xyxy
        (x1, y1, w1, h1), (x2, y2, w2, h2) = box1.chunk(4, -1), box2.chunk(4, -1)
        w1_, h1_, w2_, h2_ = w1 / 2, h1 / 2, w2 / 2, h2 / 2
        b1_x1, b1_x2, b1_y1, b1_y2 = x1 - w1_, x1 + w1_, y1 - h1_, y1 + h1_
        b2_x1, b2_x2, b2_y1, b2_y2 = x2 - w2_, x2 + w2_, y2 - h2_, y2 + h2_
    else:  # x1, y1, x2, y2 = box1
        b1_x1, b1_y1, b1_x2, b1_y2 = box1.chunk(4, -1)
        b2_x1, b2_y1, b2_x2, b2_y2 = box2.chunk(4, -1)
        w1, h1 = b1_x2 - b1_x1, b1_y2 - b1_y1 + eps
        w2, h2 = b2_x2 - b2_x1, b2_y2 - b2_y1 + eps

    # Intersection area
    inter = (b1_x2.minimum(b2_x2) - b1_x1.maximum(b2_x1)).clamp_(0) * (
        b1_y2.minimum(b2_y2) - b1_y1.maximum(b2_y1)
    ).clamp_(0)

    # Union Area
    union = w1 * h1 + w2 * h2 - inter + eps

    # IoU
    iou = inter / union
    if CIoU or DIoU or GIoU:
        cw = b1_x2.maximum(b2_x2) - b1_x1.minimum(b2_x1)  # convex (smallest enclosing box) width
        ch = b1_y2.maximum(b2_y2) - b1_y1.minimum(b2_y1)  # convex height
        if CIoU or DIoU:  # Distance or Complete IoU https://arxiv.org/abs/1911.08287v1
            c2 = cw.pow(2) + ch.pow(2) + eps  # convex diagonal squared
            rho2 = (
                (b2_x1 + b2_x2 - b1_x1 - b1_x2).pow(2) + (b2_y1 + b2_y2 - b1_y1 - b1_y2).pow(2)
            ) / 4  # center dist**2
            if CIoU:  # https://github.com/Zzh-tju/DIoU-SSD-pytorch/blob/master/utils/box/box_utils.py#L47
                v = (4 / math.pi**2) * ((w2 / h2).atan() - (w1 / h1).atan()).pow(2)
                with torch.no_grad():
                    alpha = v / (v - iou + (1 + eps))
                return iou - (rho2 / c2 + v * alpha)  # CIoU
            return iou - rho2 / c2  # DIoU
        c_area = cw * ch + eps  # convex area
        return iou - (c_area - union) / c_area  # GIoU https://arxiv.org/pdf/1902.09630.pdf
    return iou  # IoU


def _get_covariance_matrix(boxes):
    """
    Generating covariance matrix from obbs.

    Args:
        boxes (torch.Tensor): A tensor of shape (N, 5) representing rotated bounding boxes, with xywhr format.

    Returns:
        (torch.Tensor): Covariance matrices corresponding to original rotated bounding boxes.
    """
    # Gaussian bounding boxes, ignore the center points (the first two columns) because they are not needed here.
    gbbs = torch.cat((boxes[:, 2:4].pow(2) / 12, boxes[:, 4:]), dim=-1)
    a, b, c = gbbs.split(1, dim=-1)
    cos = c.cos()
    sin = c.sin()
    cos2 = cos.pow(2)
    sin2 = sin.pow(2)
    return a * cos2 + b * sin2, a * sin2 + b * cos2, (a - b) * cos * sin


def probiou(obb1, obb2, CIoU=False, eps=1e-7):
    """
    Calculate probabilistic IoU between oriented bounding boxes.

    Implements the algorithm from https://arxiv.org/pdf/2106.06072v1.pdf.

    Args:
        obb1 (torch.Tensor): Ground truth OBBs, shape (N, 5), format xywhr.
        obb2 (torch.Tensor): Predicted OBBs, shape (N, 5), format xywhr.
        CIoU (bool, optional): If True, calculate CIoU. Defaults to False.
        eps (float, optional): Small value to avoid division by zero. Defaults to 1e-7.

    Returns:
        (torch.Tensor): OBB similarities, shape (N,).

    Note:
        OBB format: [center_x, center_y, width, height, rotation_angle].
        If CIoU is True, returns CIoU instead of IoU.
    """
    x1, y1 = obb1[..., :2].split(1, dim=-1)
    x2, y2 = obb2[..., :2].split(1, dim=-1)
    a1, b1, c1 = _get_covariance_matrix(obb1)
    a2, b2, c2 = _get_covariance_matrix(obb2)

    t1 = (
        ((a1 + a2) * (y1 - y2).pow(2) + (b1 + b2) * (x1 - x2).pow(2)) / ((a1 + a2) * (b1 + b2) - (c1 + c2).pow(2) + eps)
    ) * 0.25
    t2 = (((c1 + c2) * (x2 - x1) * (y1 - y2)) / ((a1 + a2) * (b1 + b2) - (c1 + c2).pow(2) + eps)) * 0.5
    t3 = (
        ((a1 + a2) * (b1 + b2) - (c1 + c2).pow(2))
        / (4 * ((a1 * b1 - c1.pow(2)).clamp_(0) * (a2 * b2 - c2.pow(2)).clamp_(0)).sqrt() + eps)
        + eps
    ).log() * 0.5
    bd = (t1 + t2 + t3).clamp(eps, 100.0)
    hd = (1.0 - (-bd).exp() + eps).sqrt()
    iou = 1 - hd
    if CIoU:  # only include the wh aspect ratio part
        w1, h1 = obb1[..., 2:4].split(1, dim=-1)
        w2, h2 = obb2[..., 2:4].split(1, dim=-1)
        v = (4 / math.pi**2) * ((w2 / h2).atan() - (w1 / h1).atan()).pow(2)
        with torch.no_grad():
            alpha = v / (v - iou + (1 + eps))
        return iou - v * alpha  # CIoU
    return iou


def batch_probiou(obb1, obb2, eps=1e-7):
    """
    Calculate the prob IoU between oriented bounding boxes, https://arxiv.org/pdf/2106.06072v1.pdf.

    Args:
        obb1 (torch.Tensor | np.ndarray): A tensor of shape (N, 5) representing ground truth obbs, with xywhr format.
        obb2 (torch.Tensor | np.ndarray): A tensor of shape (M, 5) representing predicted obbs, with xywhr format.
        eps (float, optional): A small value to avoid division by zero. Defaults to 1e-7.

    Returns:
        (torch.Tensor): A tensor of shape (N, M) representing obb similarities.
    """
    obb1 = torch.from_numpy(obb1) if isinstance(obb1, np.ndarray) else obb1
    obb2 = torch.from_numpy(obb2) if isinstance(obb2, np.ndarray) else obb2

    x1, y1 = obb1[..., :2].split(1, dim=-1)
    x2, y2 = (x.squeeze(-1)[None] for x in obb2[..., :2].split(1, dim=-1))
    a1, b1, c1 = _get_covariance_matrix(obb1)
    a2, b2, c2 = (x.squeeze(-1)[None] for x in _get_covariance_matrix(obb2))

    t1 = (
        ((a1 + a2) * (y1 - y2).pow(2) + (b1 + b2) * (x1 - x2).pow(2)) / ((a1 + a2) * (b1 + b2) - (c1 + c2).pow(2) + eps)
    ) * 0.25
    t2 = (((c1 + c2) * (x2 - x1) * (y1 - y2)) / ((a1 + a2) * (b1 + b2) - (c1 + c2).pow(2) + eps)) * 0.5
    t3 = (
        ((a1 + a2) * (b1 + b2) - (c1 + c2).pow(2))
        / (4 * ((a1 * b1 - c1.pow(2)).clamp_(0) * (a2 * b2 - c2.pow(2)).clamp_(0)).sqrt() + eps)
        + eps
    ).log() * 0.5
    bd = (t1 + t2 + t3).clamp(eps, 100.0)
    hd = (1.0 - (-bd).exp() + eps).sqrt()
    return 1 - hd




import argparse
import os
import json
try:
    import visaionlibrary
except ImportError:
    raise ImportError("visaionlibrary not found, please check your environment")
from visaionlibrary.utils.export_onnx import export_entry

def parse_args():
    parser = argparse.ArgumentParser(description='Eval a model')
    parser.add_argument('template', help='eval config file path')
    parser.add_argument(
        '--launcher',
        choices=['none', 'pytorch', 'slurm', 'mpi'],
        default='none',
        help='job launcher')
    # When using PyTorch version >= 2.0.0, the `torch.distributed.launch`
    # will pass the `--local-rank` parameter to `tools/train.py` instead
    # of `--local_rank`.
    parser.add_argument('--local_rank', '--local-rank', type=int, default=0)
    args = parser.parse_args()
    if 'LOCAL_RANK' not in os.environ:
        os.environ['LOCAL_RANK'] = str(args.local_rank)

    return args

def export():
    args = parse_args()

    # 解析参数
    with open(args.template, 'r', encoding='utf-8') as f:
        job_info = json.load(f)

    export_entry(job_info)


if __name__ == '__main__':
    export()
   


from visaionlibrary.utils.yoloe.models.yolo.detect import DetectionPredictor
from visaionlibrary.utils.yoloe.models.yolo.segment import SegmentationPredictor
from visaionlibrary.utils.yoloe.data.augment import LetterBox, LoadVisualPrompt
from visaionlibrary.utils.yoloe.utils.instance import Instances
import numpy as np
import torch
from copy import deepcopy

class YOLOEVPPredictorMixin:
    def setup_model(self, model, verbose=True):
        """Initialize YOLO model with given parameters and set it to evaluation mode."""
        device = torch.device(self.args.device)
        self.model = model.to(device)

        self.device = device  # update device
        self.model.fp16 = False
        self.args.half = False
        self.model.eval()
        
        self.done_warmup = True
        self.return_vpe = False
        
    def set_return_vpe(self, return_vpe):
        self.return_vpe = return_vpe
    
    def set_prompts(self, prompts):
        self.prompts = deepcopy(prompts)
    
    def load_vp(self, label):
        label["img"] = label["img"].transpose(2, 0, 1)
        load_vp = LoadVisualPrompt(nc=len(label["cls"]), augment=False)
        label = load_vp(label)
        label["img"] = label["img"].transpose(1, 2, 0)
        return label
    
    def process_box_label(self, img, bboxes, cls, letterbox):
        label = dict(
            img=img,
            instances=Instances(bboxes=bboxes.astype(np.float32), 
                                segments=np.zeros((0, 1000, 2), dtype=np.float32), 
                                bbox_format="xyxy", normalized=False),
            cls=torch.tensor(cls).unsqueeze(-1)
        )
        label = letterbox(label)
        instances = label.pop("instances")
        h, w = label["img"].shape[:2]
        instances.normalize(w, h)
        instances.convert_bbox(format="xywh")
        label["bboxes"] = torch.from_numpy(instances.bboxes)
        return self.load_vp(label)
    
    def process_mask_label(self, img, masks, cls, letterbox):
        img = letterbox(image=img)
        masks = np.stack([letterbox(image=mask) for mask in masks])
        masks[masks == 114] = 0
        label = dict(
            img=img,
            masks=masks,
            cls=torch.tensor(cls).unsqueeze(-1)
        )
        return self.load_vp(label)
        
    def pre_transform(self, im):
        letterbox = LetterBox(
            self.imgsz,
            auto=False,
            stride=int(self.model.stride[-1].item()),
        )

        cls = self.prompts["cls"]
        cls = [cls] if not isinstance(cls, list) else cls
            
        if "bboxes" in self.prompts:
            bboxes = self.prompts["bboxes"]
            bboxes = [bboxes] if not isinstance(bboxes, list) else bboxes
            labels = [self.process_box_label(im[i], bboxes[i], cls[i], letterbox) for i in range(len(im))]
        elif "masks" in self.prompts:
            masks = self.prompts["masks"]
            masks = [masks] if not isinstance(masks, list) else masks
            labels = [self.process_mask_label(im[i], masks[i], cls[i], letterbox) for i in range(len(im))]
        else:
            raise ValueError("Please provide valid bboxes or masks")

        self.prompts = torch.nn.utils.rnn.pad_sequence([label["visuals"] for label in labels], batch_first=True).to(self.device)
        
        self.model.model[-1].nc = self.prompts.shape[1]
        self.model.names = [f"object{i}" for i in range(self.prompts.shape[1])]
        
        return [label["img"] for label in labels]

    def inference(self, im, *args, **kwargs):
        if self.return_vpe:
            self.vpe = self.model.get_visual_pe(im, visual=self.prompts)
        return super().inference(im, vpe=self.prompts, *args, **kwargs)


class YOLOEVPDetectPredictor(YOLOEVPPredictorMixin, DetectionPredictor):
    pass

class YOLOEVPSegPredictor(YOLOEVPPredictorMixin, SegmentationPredictor):
    pass


from pathlib import Path

from visaionlibrary.utils.yoloe.engine.model import Model  
from visaionlibrary.utils.yoloe.models import yolo
from visaionlibrary.utils.yoloe.nn.tasks import DetectionModel, SegmentationModel, YOLOEModel, YOLOESegModel


class YOLO(Model):
    """YOLO (You Only Look Once) object detection model."""

    def __init__(self, model="yolo11n.pt", task=None, verbose=False):
        """Initialize YOLO model, switching to YOLOE if model filename contains 'yoloe'."""
        path = Path(model)
        if "yoloe" in path.stem and path.suffix in {".pt", ".yaml", ".yml"}:
            new_instance = YOLOE(path, task=task, verbose=verbose)
            self.__class__ = type(new_instance)
            self.__dict__ = new_instance.__dict__
        else:
            # Continue with default YOLO initialization
            super().__init__(model=model, task=task, verbose=verbose)

    @property
    def task_map(self):
        """Map head to model and predictor classes."""
        return {
            "detect": {
                "model": DetectionModel,
                "predictor": yolo.detect.DetectionPredictor,
            },
            "segment": {
                "model": SegmentationModel,
                "predictor": yolo.segment.SegmentationPredictor,
            },
        }


class YOLOE(Model):
    """YOLOE object detection and segmentation model."""

    def __init__(self, model="yoloe-v8s-seg.pt", task=None, verbose=False) -> None:
        """
        Initialize YOLOE model with a pre-trained model file.

        Args:
            model (str | Path): Path to the pre-trained model file. Supports *.pt and *.yaml formats.
            verbose (bool): If True, prints additional information during initialization.
        """
        super().__init__(model=model, task=task, verbose=verbose)

    @property
    def task_map(self):
        """Map head to model and predictor classes."""
        return {
            "detect": {
                "model": YOLOEModel,
                "predictor": yolo.detect.DetectionPredictor,
            },
            "segment": {
                "model": YOLOESegModel,
                "predictor": yolo.segment.SegmentationPredictor,
            },
        }

    def get_text_pe(self, texts):
        assert(isinstance(self.model, YOLOEModel))
        return self.model.get_text_pe(texts)
    
    def get_visual_pe(self, img, visual):
        assert(isinstance(self.model, YOLOEModel))
        return self.model.get_visual_pe(img, visual)

    def set_vocab(self, vocab, names):
        assert(isinstance(self.model, YOLOEModel))
        self.model.set_vocab(vocab, names=names)
    
    def get_vocab(self, names):
        assert(isinstance(self.model, YOLOEModel))
        return self.model.get_vocab(names)

    def set_classes(self, classes, embeddings):
        """
        Set classes.

        Args:
            classes (List(str)): A list of categories i.e. ["person"].
        """
        assert(isinstance(self.model, YOLOEModel))
        self.model.set_classes(classes, embeddings)
        # Remove background if it's given
        assert(" " not in classes)
        self.model.names = classes

        # Reset method class names
        # self.predictor = None  # reset predictor otherwise old names remain
        if self.predictor:
            self.predictor.model.names = classes



import argparse
import logging
import os
import platform
import os.path as osp
import json
from typing import Union, List, Dict, Any

from mmdet.utils import setup_cache_size_limit_of_dynamo
from mmengine.config import Config, DictAction
from mmengine.logging import print_log
from mmengine.runner import Runner
from mmengine.registry import RUNNERS
from mmyolo.utils import is_metainfo_lower

try:
    import visaionlibrary
except ImportError:
    raise ImportError("visaionlibrary not found, please check your environment")

def trigger_det_visualization_hook(cfg):
    default_hooks = cfg.default_hooks
    if 'visualization' in default_hooks:
        visualization_hook = default_hooks['visualization']
        visualization_hook['draw'] = True
        visualization_hook['test_out_dir'] = "show_dir"
    else:
        raise RuntimeError(
            'VisualizationHook must be included in default_hooks.'
            'refer to usage '
            '"visualization=dict(type=\'VisualizationHook\')"')

    return cfg

def trigger_seg_visualization_hook(cfg):
    default_hooks = cfg.default_hooks
    if 'visualization' in default_hooks:
        visualization_hook = default_hooks['visualization']
        visualization_hook['draw'] = True
        visualization_hook['interval'] = 1
    else:
        raise RuntimeError(
            'VisualizationHook must be included in default_hooks.'
            'refer to usage '
            '"visualization=dict(type=\'VisualizationHook\')"')

    return cfg

def trigger_cls_visualization_hook(cfg, show_dir):
    default_hooks = cfg.default_hooks
    if 'visualization' in default_hooks:
        visualization_hook = default_hooks['visualization']
        visualization_hook['out_dir'] = os.path.join(show_dir, "show_dir")
        visualization_hook['interval'] = 1
        visualization_hook['enable'] = True
    return cfg

def _extract_baseParam_visaion_keys(data, result=None):
    """
    递归提取以 VISAION 开头的 key 和 value
    """
    if result is None:
        result = {}
    if isinstance(data, dict):
        for key, value in data.items():
            if key.startswith('VISAION'):
                result[key] = value
            _extract_baseParam_visaion_keys(value, result)
    elif isinstance(data, list):
        for item in data:
            _extract_baseParam_visaion_keys(item, result)
    return result

def parse_args():
    parser = argparse.ArgumentParser(description='Eval a model')
    parser.add_argument('template', help='eval config file path')
    parser.add_argument(
        '--launcher',
        choices=['none', 'pytorch', 'slurm', 'mpi'],
        default='none',
        help='job launcher')
    # When using PyTorch version >= 2.0.0, the `torch.distributed.launch`
    # will pass the `--local-rank` parameter to `tools/train.py` instead
    # of `--local_rank`.
    parser.add_argument('--local_rank', '--local-rank', type=int, default=0)
    args = parser.parse_args()
    if 'LOCAL_RANK' not in os.environ:
        os.environ['LOCAL_RANK'] = str(args.local_rank)

    return args


def eval():
    args = parse_args()

    # 解析参数
    with open(args.template, 'r', encoding='utf-8') as f:
        job_info = json.load(f)

    # 提取project_id, job_id, job_name, job_args
    project_id = job_info['project_id']     # 项目id
    job_id = job_info['job_id']             # 任务id
    job_type = job_info['job_type']         # 任务名称
    job_args = job_info['job_args']           # 任务参数
    assert project_id is not None, "project_id is not None"
    assert job_id is not None, "job_id is not None"
    assert job_type is not None, "job_type is not None"
    assert job_args is not None, "job_args is not None"

    # 提取以 VISAION 开头的 key 和 value
    visaion_params = _extract_baseParam_visaion_keys(job_args['baseParam'])  # 提取基础参数
    #  打印 VISAION 参数, 设置环境变量
    print_log('setting environment variables...')
    for k, v in visaion_params.items():
        v = str(v)
        # windows 下使用\\加载config时会失败, 所以需要替换为/
        if platform.system() == 'Windows':
            v = v.replace('\\', '/')
        print_log(f"{k}={v}")
        os.environ[k] = v
    
    # 根据参数, 获取配置文件路径, 模型文件路径, 保存eval的文件夹路径, 任务类型
    VISAION_DIR = visaion_params['VISAION_DIR']
    task_type = visaion_params['VISAION_TASK_TYPE']
    config_path = osp.join(VISAION_DIR,visaion_params['VISAION_CONFIG_PATH'])               # 配置文件路径, py文件, 训练后存储的配置文件
    checkpoint_path = osp.join(VISAION_DIR, visaion_params['VISAION_PTH_PATH'])             # 训练后保存的模型文件
    work_dir = osp.join(VISAION_DIR, "projects", str(project_id), "eval", str(job_id)).replace('\\', '/')      # 保存eval的文件夹路径
    data_root = visaion_params['VISAION_TEST_DATA_ROOT']    # 测试数据集路径
    test_data_root = [osp.join(VISAION_DIR, "projects",str(project_id), "datasets", data_i) for data_i in data_root]
    test_data_root = "::".join(test_data_root)
    assert osp.exists(config_path), f'config file {config_path} not found'
    assert osp.exists(checkpoint_path), f'checkpoint file {checkpoint_path} not found'
    assert work_dir is not None, f'work_dir is None'
    assert task_type in ['det', 'seg', 'cls', 'ins'], f'task_type {task_type} not supported'

    # Reduce the number of repeated compilations and improve
    # training speed.
    setup_cache_size_limit_of_dynamo()

    # load config
    cfg = Config.fromfile(config_path)
    cfg.load_from = checkpoint_path  # 设置模型文件路径
    cfg.work_dir = work_dir  # 设置保存eval的文件夹路径
    cfg.launcher = args.launcher  # 设置任务类型
    cfg.test_dataloader.dataset.data_root = test_data_root  # 设置测试数据集路径
    os.environ["VISAION_WORK_DIR"] = work_dir   # set env variable, used in some hooks, metrics, etc.
    if task_type == 'det':
        trigger_det_visualization_hook(cfg)
    elif task_type == 'seg':
        trigger_seg_visualization_hook(cfg)
    elif task_type == 'cls':
        trigger_cls_visualization_hook(cfg, work_dir)
    elif task_type == 'ins':
        trigger_det_visualization_hook(cfg)
    else:
        raise ValueError(f"task_type {task_type} not supported")

    
    # 打印visaionlibrary版本
    print_log(f"visaionlibrary version: {visaionlibrary.__version__}")
    
    # Determine whether the custom metainfo fields are all lowercase
    is_metainfo_lower(cfg)

    # build the runner from config
    if 'runner_type' not in cfg:
        # build the default runner
        runner = Runner.from_cfg(cfg)
    else:
        # build customized runner from the registry
        # if 'runner_type' is set in the cfg
        runner = RUNNERS.build(cfg)

    # start testing
    runner.test()

    runner.logger.info("VISAION_DONE")

if __name__ == '__main__':
    eval()



from typing import Iterator, List, Union, Iterable
from torch.utils.data import BatchSampler, Sampler
from mmengine.registry import DATA_SAMPLERS

from visaionlibrary.datasets.samplers.sampler import InfiniteSamplerBatch

T_Sampler = Union[
    InfiniteSamplerBatch,
]

@DATA_SAMPLERS.register_module()
class BatchSamplerConstantOne(BatchSampler):
    """
    BatchSampler的子类, 强行将batch_size=1, 因为迭代 sampler(例如InfiniteSamplerBatch) 返回的是一个list(这个list是一个batchsize的数据), 而不是一个int, 且
    所以强行让batch_size=1, 然后在__iter__中, 将list(这个list是一个batchsize的数据)的数据展开成一个batch,
    所以batch_size=1, 然后drop_last=False, 这两个参数废弃
    """
    def __init__(self, sampler: T_Sampler, batch_size: int = 1, drop_last: bool = False) -> None:
        super().__init__(sampler=sampler, batch_size=1, drop_last=False)

    def __iter__(self) -> Iterator[List[int]]:
        batch: List[List[int]] = []
        idx: List[int]
        for idx in self.sampler:    # 迭代sampler, 返回的是一个list(这个list是一个batchsize的数据), 而不是一个int, idx 例如[46, 49, 34, 0, 31, 22, 47, 11, 40, 17, 17, 43], 负样本在后面
            batch.append(idx)
            if len(batch) == self.batch_size:   # batchsize=1, 所以len(batch)=1, 所以len(batch) == self.batch_size, 所以yield batch
                iter_list = [item for sublist in batch for item in sublist]     # batch=[[46, 49, 34, 0, 31, 22, 47, 11, 40, 17, 17, 43]]
                yield iter_list
                batch = []
        # if len(batch) > 0 and not self.drop_last:
        #     yield batch

    def __len__(self) -> int:
        return len(self.sampler)



import re
import numpy as np


def decode_str2mask(mask_str: str, height: int, width: int) -> np.array:
        """
        解压str 到 array, web为了方便传输, 将mask压缩成字符串, 这里解压
        :param mask_str:
        :param height:
        :param width:
        :return:
        """
        _rleDecodeMap = {'Z': '0', 'N': '1'}  # decode 对应map

        if mask_str == "":
            return None

        # groups all encoded pieces together
        mask_list = re.findall(r'(\d+)(\w|\s)', mask_str, re.S) # [('449', 'Z'), ('12', 'N'), ('153', 'Z'), ....]

        # repeat each piece's last char with number
        repeat_list = list(map(lambda x: _rleDecodeMap[x[-1]] * int(x[0]), mask_list))

        # connect && to list
        np_flatten = [int(i) for i in ''.join(repeat_list)]

        # list->mask
        mask = np.reshape(np_flatten, (height, width))
        mask = mask.astype(np.uint8)
        mask = np.ascontiguousarray(mask)
        return mask

def encode_mask_to_str(mask: np.array) -> str:
    """
    将mask编码为字符串
    :param mask: 二值mask数组
    :return: 编码后的字符串
    """
    _rleEncodeMap = {'0': 'Z', '1': 'N'}  # encode 对应map
    
    # 将mask展平为一维数组
    mask_flat = mask.flatten()
    
    # 转换为字符串
    mask_str = ''.join(map(str, mask_flat))
    
    # 使用正则表达式找出连续的相同数字
    pattern = r'(0+|1+)'
    matches = re.finditer(pattern, mask_str)
    
    # 编码结果
    encoded = []
    for match in matches:
        group = match.group()
        count = len(group)
        value = group[0]
        encoded.append(f"{count}{_rleEncodeMap[value]}")
    
    return ''.join(encoded)


import math
import torch
import torch.nn as nn
from mmengine.model import BaseModule


class Proto(nn.Module):
    """Ultralytics YOLO models mask Proto module for segmentation models."""

    def __init__(self, c1: int, c_: int = 256, c2: int = 32):
        """Initialize the Ultralytics YOLO models mask Proto module with specified number of protos and masks.

        Args:
            c1 (int): Input channels.
            c_ (int): Intermediate channels.
            c2 (int): Output channels (number of protos).
        """
        super().__init__()
        self.cv1 = Conv(c1, c_, k=3)
        self.upsample = nn.ConvTranspose2d(c_, c_, 2, 2, 0, bias=True)  # nn.Upsample(scale_factor=2, mode='nearest')
        self.cv2 = Conv(c_, c_, k=3)
        self.cv3 = Conv(c_, c2)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Perform a forward pass through layers using an upsampled input image."""
        return self.cv3(self.cv2(self.upsample(self.cv1(x))))

def autopad(k, p=None, d=1):  # kernel, padding, dilation
    """Pad to 'same' shape outputs."""
    if d > 1:
        k = d * (k - 1) + 1 if isinstance(k, int) else [d * (x - 1) + 1 for x in k]  # actual kernel-size
    if p is None:
        p = k // 2 if isinstance(k, int) else [x // 2 for x in k]  # auto-pad
    return p

class Concat(nn.Module):
    """Concatenate a list of tensors along specified dimension.

    Attributes:
        d (int): Dimension along which to concatenate tensors.
    """

    def __init__(self, dimension=1):
        """Initialize Concat module.

        Args:
            dimension (int): Dimension along which to concatenate tensors.
        """
        super().__init__()
        self.d = dimension

    def forward(self, x: list[torch.Tensor]):
        """Concatenate input tensors along specified dimension.

        Args:
            x (list[torch.Tensor]): List of input tensors.

        Returns:
            (torch.Tensor): Concatenated tensor.
        """
        return torch.cat(x, self.d)

class Conv(nn.Module):
    """Standard convolution module with batch normalization and activation.

    Attributes:
        conv (nn.Conv2d): Convolutional layer.
        bn (nn.BatchNorm2d): Batch normalization layer.
        act (nn.Module): Activation function layer.
        default_act (nn.Module): Default activation function (SiLU).
    """

    default_act = nn.SiLU()  # default activation

    def __init__(self, c1, c2, k=1, s=1, p=None, g=1, d=1, act=True):
        """Initialize Conv layer with given parameters.

        Args:
            c1 (int): Number of input channels.
            c2 (int): Number of output channels.
            k (int): Kernel size.
            s (int): Stride.
            p (int, optional): Padding.
            g (int): Groups.
            d (int): Dilation.
            act (bool | nn.Module): Activation function.
        """
        super().__init__()
        self.conv = nn.Conv2d(c1, c2, k, s, autopad(k, p, d), groups=g, dilation=d, bias=False)
        self.bn = nn.BatchNorm2d(c2)
        self.act = self.default_act if act is True else act if isinstance(act, nn.Module) else nn.Identity()

    def forward(self, x):
        """Apply convolution, batch normalization and activation to input tensor.

        Args:
            x (torch.Tensor): Input tensor.

        Returns:
            (torch.Tensor): Output tensor.
        """
        return self.act(self.bn(self.conv(x)))

    def forward_fuse(self, x):
        """Apply convolution and activation without batch normalization.

        Args:
            x (torch.Tensor): Input tensor.

        Returns:
            (torch.Tensor): Output tensor.
        """
        return self.act(self.conv(x))

class DWConv(Conv):
    """Depth-wise convolution module."""

    def __init__(self, c1, c2, k=1, s=1, d=1, act=True):
        """Initialize depth-wise convolution with given parameters.

        Args:
            c1 (int): Number of input channels.
            c2 (int): Number of output channels.
            k (int): Kernel size.
            s (int): Stride.
            d (int): Dilation.
            act (bool | nn.Module): Activation function.
        """
        super().__init__(c1, c2, k, s, g=math.gcd(c1, c2), d=d, act=act)

class Attention(nn.Module):
    """Attention module that performs self-attention on the input tensor.

    Args:
        dim (int): The input tensor dimension.
        num_heads (int): The number of attention heads.
        attn_ratio (float): The ratio of the attention key dimension to the head dimension.

    Attributes:
        num_heads (int): The number of attention heads.
        head_dim (int): The dimension of each attention head.
        key_dim (int): The dimension of the attention key.
        scale (float): The scaling factor for the attention scores.
        qkv (Conv): Convolutional layer for computing the query, key, and value.
        proj (Conv): Convolutional layer for projecting the attended values.
        pe (Conv): Convolutional layer for positional encoding.
    """

    def __init__(self, dim: int, num_heads: int = 8, attn_ratio: float = 0.5):
        """Initialize multi-head attention module.

        Args:
            dim (int): Input dimension.
            num_heads (int): Number of attention heads.
            attn_ratio (float): Attention ratio for key dimension.
        """
        super().__init__()
        self.num_heads = num_heads
        self.head_dim = dim // num_heads
        self.key_dim = int(self.head_dim * attn_ratio)
        self.scale = self.key_dim**-0.5
        nh_kd = self.key_dim * num_heads
        h = dim + nh_kd * 2
        self.qkv = Conv(dim, h, 1, act=False)
        self.proj = Conv(dim, dim, 1, act=False)
        self.pe = Conv(dim, dim, 3, 1, g=dim, act=False)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of the Attention module.

        Args:
            x (torch.Tensor): The input tensor.

        Returns:
            (torch.Tensor): The output tensor after self-attention.
        """
        B, C, H, W = x.shape
        N = H * W
        qkv = self.qkv(x)
        q, k, v = qkv.view(B, self.num_heads, self.key_dim * 2 + self.head_dim, N).split(
            [self.key_dim, self.key_dim, self.head_dim], dim=2
        )

        attn = (q.transpose(-2, -1) @ k) * self.scale
        attn = attn.softmax(dim=-1)
        x = (v @ attn.transpose(-2, -1)).view(B, C, H, W) + self.pe(v.reshape(B, C, H, W))
        x = self.proj(x)
        return x

class PSABlock(nn.Module):
    """PSABlock class implementing a Position-Sensitive Attention block for neural networks.

    This class encapsulates the functionality for applying multi-head attention and feed-forward neural network layers
    with optional shortcut connections.

    Attributes:
        attn (Attention): Multi-head attention module.
        ffn (nn.Sequential): Feed-forward neural network module.
        add (bool): Flag indicating whether to add shortcut connections.

    Methods:
        forward: Performs a forward pass through the PSABlock, applying attention and feed-forward layers.

    Examples:
        Create a PSABlock and perform a forward pass
        >>> psablock = PSABlock(c=128, attn_ratio=0.5, num_heads=4, shortcut=True)
        >>> input_tensor = torch.randn(1, 128, 32, 32)
        >>> output_tensor = psablock(input_tensor)
    """

    def __init__(self, c: int, attn_ratio: float = 0.5, num_heads: int = 4, shortcut: bool = True) -> None:
        """Initialize the PSABlock.

        Args:
            c (int): Input and output channels.
            attn_ratio (float): Attention ratio for key dimension.
            num_heads (int): Number of attention heads.
            shortcut (bool): Whether to use shortcut connections.
        """
        super().__init__()

        self.attn = Attention(c, attn_ratio=attn_ratio, num_heads=num_heads)
        self.ffn = nn.Sequential(Conv(c, c * 2, 1), Conv(c * 2, c, 1, act=False))
        self.add = shortcut

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Execute a forward pass through PSABlock.

        Args:
            x (torch.Tensor): Input tensor.

        Returns:
            (torch.Tensor): Output tensor after attention and feed-forward processing.
        """
        x = x + self.attn(x) if self.add else self.attn(x)
        x = x + self.ffn(x) if self.add else self.ffn(x)
        return x

class C2PSA(nn.Module):
    """C2PSA module with attention mechanism for enhanced feature extraction and processing.

    This module implements a convolutional block with attention mechanisms to enhance feature extraction and processing
    capabilities. It includes a series of PSABlock modules for self-attention and feed-forward operations.

    Attributes:
        c (int): Number of hidden channels.
        cv1 (Conv): 1x1 convolution layer to reduce the number of input channels to 2*c.
        cv2 (Conv): 1x1 convolution layer to reduce the number of output channels to c.
        m (nn.Sequential): Sequential container of PSABlock modules for attention and feed-forward operations.

    Methods:
        forward: Performs a forward pass through the C2PSA module, applying attention and feed-forward operations.

    Examples:
        >>> c2psa = C2PSA(c1=256, c2=256, n=3, e=0.5)
        >>> input_tensor = torch.randn(1, 256, 64, 64)
        >>> output_tensor = c2psa(input_tensor)

    Notes:
        This module essentially is the same as PSA module, but refactored to allow stacking more PSABlock modules.
    """

    def __init__(self, c1: int, c2: int, n: int = 1, e: float = 0.5):
        """Initialize C2PSA module.

        Args:
            c1 (int): Input channels.
            c2 (int): Output channels.
            n (int): Number of PSABlock modules.
            e (float): Expansion ratio.
        """
        super().__init__()
        assert c1 == c2
        self.c = int(c1 * e)
        self.cv1 = Conv(c1, 2 * self.c, 1, 1)
        self.cv2 = Conv(2 * self.c, c1, 1)

        self.m = nn.Sequential(*(PSABlock(self.c, attn_ratio=0.5, num_heads=self.c // 64) for _ in range(n)))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Process the input tensor through a series of PSA blocks.

        Args:
            x (torch.Tensor): Input tensor.

        Returns:
            (torch.Tensor): Output tensor after processing.
        """
        a, b = self.cv1(x).split((self.c, self.c), dim=1)
        b = self.m(b)
        return self.cv2(torch.cat((a, b), 1))

class Bottleneck(nn.Module):
    """Standard bottleneck."""

    def __init__(
        self, c1: int, c2: int, shortcut: bool = True, g: int = 1, k: tuple[int, int] = (3, 3), e: float = 0.5
    ):
        """Initialize a standard bottleneck module.

        Args:
            c1 (int): Input channels.
            c2 (int): Output channels.
            shortcut (bool): Whether to use shortcut connection.
            g (int): Groups for convolutions.
            k (tuple): Kernel sizes for convolutions.
            e (float): Expansion ratio.
        """
        super().__init__()
        c_ = int(c2 * e)  # hidden channels
        self.cv1 = Conv(c1, c_, k[0], 1)
        self.cv2 = Conv(c_, c2, k[1], 1, g=g)
        self.add = shortcut and c1 == c2

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Apply bottleneck with optional shortcut connection."""
        return x + self.cv2(self.cv1(x)) if self.add else self.cv2(self.cv1(x))

class C3(nn.Module):
    """CSP Bottleneck with 3 convolutions."""

    def __init__(self, c1: int, c2: int, n: int = 1, shortcut: bool = True, g: int = 1, e: float = 0.5):
        """Initialize the CSP Bottleneck with 3 convolutions.

        Args:
            c1 (int): Input channels.
            c2 (int): Output channels.
            n (int): Number of Bottleneck blocks.
            shortcut (bool): Whether to use shortcut connections.
            g (int): Groups for convolutions.
            e (float): Expansion ratio.
        """
        super().__init__()
        c_ = int(c2 * e)  # hidden channels
        self.cv1 = Conv(c1, c_, 1, 1)
        self.cv2 = Conv(c1, c_, 1, 1)
        self.cv3 = Conv(2 * c_, c2, 1)  # optional act=FReLU(c2)
        self.m = nn.Sequential(*(Bottleneck(c_, c_, shortcut, g, k=((1, 1), (3, 3)), e=1.0) for _ in range(n)))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass through the CSP bottleneck with 3 convolutions."""
        return self.cv3(torch.cat((self.m(self.cv1(x)), self.cv2(x)), 1))

class C3k(C3):
    """C3k is a CSP bottleneck module with customizable kernel sizes for feature extraction in neural networks."""

    def __init__(self, c1: int, c2: int, n: int = 1, shortcut: bool = True, g: int = 1, e: float = 0.5, k: int = 3):
        """Initialize C3k module.

        Args:
            c1 (int): Input channels.
            c2 (int): Output channels.
            n (int): Number of Bottleneck blocks.
            shortcut (bool): Whether to use shortcut connections.
            g (int): Groups for convolutions.
            e (float): Expansion ratio.
            k (int): Kernel size.
        """
        super().__init__(c1, c2, n, shortcut, g, e)
        c_ = int(c2 * e)  # hidden channels
        # self.m = nn.Sequential(*(RepBottleneck(c_, c_, shortcut, g, k=(k, k), e=1.0) for _ in range(n)))
        self.m = nn.Sequential(*(Bottleneck(c_, c_, shortcut, g, k=(k, k), e=1.0) for _ in range(n)))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass through the CSP bottleneck with 3 convolutions."""
        return self.cv3(torch.cat((self.m(self.cv1(x)), self.cv2(x)), 1))

class C2f(nn.Module):
    """Faster Implementation of CSP Bottleneck with 2 convolutions."""

    def __init__(self, c1: int, c2: int, n: int = 1, shortcut: bool = False, g: int = 1, e: float = 0.5):
        """Initialize a CSP bottleneck with 2 convolutions.

        Args:
            c1 (int): Input channels.
            c2 (int): Output channels.
            n (int): Number of Bottleneck blocks.
            shortcut (bool): Whether to use shortcut connections.
            g (int): Groups for convolutions.
            e (float): Expansion ratio.
        """
        super().__init__()
        self.c = int(c2 * e)  # hidden channels
        self.cv1 = Conv(c1, 2 * self.c, 1, 1)
        self.cv2 = Conv((2 + n) * self.c, c2, 1)  # optional act=FReLU(c2)
        self.m = nn.ModuleList(Bottleneck(self.c, self.c, shortcut, g, k=((3, 3), (3, 3)), e=1.0) for _ in range(n))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass through C2f layer."""
        y = list(self.cv1(x).chunk(2, 1))
        y.extend(m(y[-1]) for m in self.m)
        return self.cv2(torch.cat(y, 1))

    def forward_split(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass using split() instead of chunk()."""
        y = self.cv1(x).split((self.c, self.c), 1)
        y = [y[0], y[1]]
        y.extend(m(y[-1]) for m in self.m)
        return self.cv2(torch.cat(y, 1))


class C3K2(C2f):  # C2f
    def __init__(
            self,
            in_channels: int,
            out_channels: int,
            expand_ratio: float = 0.5,
            num_blocks: int = 2,
            add_identity: bool = True,  # shortcut
            c3k: bool = False,  # c3k
            conv_groups: int = 1) -> None:
        super().__init__(in_channels, out_channels, num_blocks, add_identity, conv_groups, expand_ratio)
        self.m = nn.ModuleList(
            C3k(self.c, self.c, 2, add_identity, conv_groups) if c3k else Bottleneck(self.c, self.c, add_identity, conv_groups) for _ in range(num_blocks)
        )


class SPPF(nn.Module):
    """Spatial Pyramid Pooling - Fast (SPPF) layer for YOLOv5 by Glenn Jocher."""

    def __init__(self, c1: int, c2: int, k: int = 5):
        """Initialize the SPPF layer with given input/output channels and kernel size.

        Args:
            c1 (int): Input channels.
            c2 (int): Output channels.
            k (int): Kernel size.

        Notes:
            This module is equivalent to SPP(k=(5, 9, 13)).
        """
        super().__init__()
        c_ = c1 // 2  # hidden channels
        self.cv1 = Conv(c1, c_, 1, 1)
        self.cv2 = Conv(c_ * 4, c2, 1, 1)
        self.m = nn.MaxPool2d(kernel_size=k, stride=1, padding=k // 2)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Apply sequential pooling operations to input and return concatenated feature maps."""
        y = [self.cv1(x)]
        y.extend(self.m(y[-1]) for _ in range(3))
        return self.cv2(torch.cat(y, 1))



import contextlib
import pickle
import re
import types
from copy import deepcopy
from pathlib import Path

import torch
import torch.nn as nn

from visaionlibrary.utils.yoloe.nn.modules import (
    AIFI,
    C1,
    C2,
    C2PSA,
    C3,
    C3TR,
    ELAN1,
    OBB,
    PSA,
    SPP,
    SPPELAN,
    SPPF,
    AConv,
    ADown,
    Bottleneck,
    BottleneckCSP,
    C2f,
    C2fAttn,
    C2fCIB,
    C2fPSA,
    C3Ghost,
    C3k2,
    C3x,
    CBFuse,
    CBLinear,
    Classify,
    Concat,
    Conv,
    Conv2,
    ConvTranspose,
    Detect,
    DWConv,
    DWConvTranspose2d,
    Focus,
    GhostBottleneck,
    GhostConv,
    HGBlock,
    HGStem,
    ImagePoolingAttn,
    Pose,
    RepC3,
    RepConv,
    RepNCSPELAN4,
    RepVGGDW,
    ResNetLayer,
    RTDETRDecoder,
    SCDown,
    Segment,
    YOLOEDetect,
    YOLOESegment,
    v10Detect,
    LRPCHead,
    MaxSigmoidAttnBlock,
)
from visaionlibrary.utils.yoloe.utils import yaml_load
from visaionlibrary.utils.yoloe.utils.checks import check_yaml
from visaionlibrary.utils.yoloe.utils.loss import (
    E2EDetectLoss,
    v8DetectionLoss,
    v8SegmentationLoss,
    TVPDetectLoss,
    TVPSegmentLoss
)
from visaionlibrary.utils.yoloe.utils.ops import make_divisible
from visaionlibrary.utils.yoloe.utils.torch_utils import (
    fuse_conv_and_bn,
    fuse_deconv_and_bn,
    initialize_weights,
    intersect_dicts,
    scale_img,
    time_sync,
    smart_inference_mode
)
from visaionlibrary.utils.yoloe.nn.autobackend import check_class_names
from visaionlibrary.utils.yoloe.utils import DEFAULT_CFG_DICT, DEFAULT_CFG_KEYS


class BaseModel(nn.Module):
    """The BaseModel class serves as a base class for all the models."""

    def forward(self, x, *args, **kwargs):
        """
        Perform forward pass of the model for either training or inference.

        If x is a dict, calculates and returns the loss for training. Otherwise, returns predictions for inference.

        Args:
            x (torch.Tensor | dict): Input tensor for inference, or dict with image tensor and labels for training.
            *args (Any): Variable length argument list.
            **kwargs (Any): Arbitrary keyword arguments.

        Returns:
            (torch.Tensor): Loss if x is a dict (training), or network predictions (inference).
        """
        if isinstance(x, dict):  # for cases of training and validating while training.
            return self.loss(x, *args, **kwargs)
        return self.predict(x, *args, **kwargs)

    def predict(self, x, profile=False, visualize=False, augment=False, embed=None):
        """
        Perform a forward pass through the network.

        Args:
            x (torch.Tensor): The input tensor to the model.
            profile (bool):  Print the computation time of each layer if True, defaults to False.
            visualize (bool): Save the feature maps of the model if True, defaults to False.
            augment (bool): Augment image during prediction, defaults to False.
            embed (list, optional): A list of feature vectors/embeddings to return.

        Returns:
            (torch.Tensor): The last output of the model.
        """
        if augment:
            return self._predict_augment(x)
        return self._predict_once(x, profile, visualize, embed)

    def _predict_once(self, x, profile=False, visualize=False, embed=None):
        """
        Perform a forward pass through the network.

        Args:
            x (torch.Tensor): The input tensor to the model.
            profile (bool):  Print the computation time of each layer if True, defaults to False.
            visualize (bool): Save the feature maps of the model if True, defaults to False.
            embed (list, optional): A list of feature vectors/embeddings to return.

        Returns:
            (torch.Tensor): The last output of the model.
        """
        y, dt, embeddings = [], [], []  # outputs
        for m in self.model:
            if m.f != -1:  # if not from previous layer
                x = y[m.f] if isinstance(m.f, int) else [x if j == -1 else y[j] for j in m.f]  # from earlier layers
            if profile:
                self._profile_one_layer(m, x, dt)
            x = m(x)  # run
            y.append(x if m.i in self.save else None)  # save output

            if embed and m.i in embed:
                embeddings.append(nn.functional.adaptive_avg_pool2d(x, (1, 1)).squeeze(-1).squeeze(-1))  # flatten
                if m.i == max(embed):
                    return torch.unbind(torch.cat(embeddings, 1), dim=0)
        return x

    def _predict_augment(self, x):
        """Perform augmentations on input image x and return augmented inference."""
        return self._predict_once(x)

    def _profile_one_layer(self, m, x, dt):
        """
        Profile the computation time and FLOPs of a single layer of the model on a given input. Appends the results to
        the provided list.

        Args:
            m (nn.Module): The layer to be profiled.
            x (torch.Tensor): The input data to the layer.
            dt (list): A list to store the computation time of the layer.

        Returns:
            None
        """
        c = m == self.model[-1] and isinstance(x, list)  # is final layer list, copy input as inplace fix
        t = time_sync()
        for _ in range(10):
            m(x.copy() if c else x)
        dt.append((time_sync() - t) * 100)

    def fuse(self, verbose=True):
        """
        Fuse the `Conv2d()` and `BatchNorm2d()` layers of the model into a single layer, in order to improve the
        computation efficiency.

        Returns:
            (nn.Module): The fused model is returned.
        """
        if not self.is_fused():
            for m in self.model.modules():
                if isinstance(m, (Conv, Conv2, DWConv)) and hasattr(m, "bn"):
                    if isinstance(m, Conv2):
                        m.fuse_convs()
                    m.conv = fuse_conv_and_bn(m.conv, m.bn)  # update conv
                    delattr(m, "bn")  # remove batchnorm
                    m.forward = m.forward_fuse  # update forward
                if isinstance(m, ConvTranspose) and hasattr(m, "bn"):
                    m.conv_transpose = fuse_deconv_and_bn(m.conv_transpose, m.bn)
                    delattr(m, "bn")  # remove batchnorm
                    m.forward = m.forward_fuse  # update forward
                if isinstance(m, RepConv):
                    m.fuse_convs()
                    m.forward = m.forward_fuse  # update forward
                if isinstance(m, RepVGGDW):
                    m.fuse()
                    m.forward = m.forward_fuse
                device = next(self.model.parameters()).device
                if isinstance(m, MaxSigmoidAttnBlock):
                    assert(isinstance(self, YOLOEModel))
                    m.fuse(self.pe.to(device))
                if isinstance(m, YOLOEDetect) and hasattr(self, 'pe'):
                    assert(isinstance(self, YOLOEModel))
                    m.fuse(self.pe.to(device))
            self.info(verbose=verbose)

        return self

    def is_fused(self, thresh=10):
        """
        Check if the model has less than a certain threshold of BatchNorm layers.

        Args:
            thresh (int, optional): The threshold number of BatchNorm layers. Default is 10.

        Returns:
            (bool): True if the number of BatchNorm layers in the model is less than the threshold, False otherwise.
        """
        bn = tuple(v for k, v in nn.__dict__.items() if "Norm" in k)  # normalization layers, i.e. BatchNorm2d()
        return sum(isinstance(v, bn) for v in self.modules()) < thresh  # True if < 'thresh' BatchNorm layers in model

    def info(self, detailed=False, verbose=True, imgsz=640):
        """
        Prints model information.

        Args:
            detailed (bool): if True, prints out detailed information about the model. Defaults to False
            verbose (bool): if True, prints out the model information. Defaults to False
            imgsz (int): the size of the image that the model will be trained on. Defaults to 640
        """
        return None

    def _apply(self, fn):
        """
        Applies a function to all the tensors in the model that are not parameters or registered buffers.

        Args:
            fn (function): the function to apply to the model

        Returns:
            (BaseModel): An updated BaseModel object.
        """
        self = super()._apply(fn)
        m = self.model[-1]  # Detect()
        if isinstance(m, Detect):  # includes all Detect subclasses like Segment, Pose, OBB, YOLOEDetect, YOLOESegment
            m.stride = fn(m.stride)
            m.anchors = fn(m.anchors)
            m.strides = fn(m.strides)
        return self

    def load(self, weights, verbose=True):
        """
        Load the weights into the model.

        Args:
            weights (dict | torch.nn.Module): The pre-trained weights to be loaded.
            verbose (bool, optional): Whether to log the transfer progress. Defaults to True.
        """
        model = weights["model"] if isinstance(weights, dict) else weights  # torchvision models are not dicts
        csd = model.float().state_dict()  # checkpoint state_dict as FP32
        csd = intersect_dicts(csd, self.state_dict())  # intersect
        self.load_state_dict(csd, strict=False)  # load

    def loss(self, batch, preds=None):
        """
        Compute loss.

        Args:
            batch (dict): Batch to compute loss on
            preds (torch.Tensor | List[torch.Tensor]): Predictions.
        """
        if getattr(self, "criterion", None) is None:
            self.criterion = self.init_criterion()

        preds = self.forward(batch["img"]) if preds is None else preds
        return self.criterion(preds, batch)

    def init_criterion(self):
        """Initialize the loss criterion for the BaseModel."""
        raise NotImplementedError("compute_loss() needs to be implemented by task heads")


class DetectionModel(BaseModel):
    """YOLOv8 detection model."""

    def __init__(self, cfg="yolov8n.yaml", ch=3, nc=None, verbose=True):  # model, input channels, number of classes
        """Initialize the YOLOv8 detection model with the given config and parameters."""
        super().__init__()
        self.yaml = cfg if isinstance(cfg, dict) else yaml_model_load(cfg)  # cfg dict
        if self.yaml["backbone"][0][2] == "Silence":
            self.yaml["backbone"][0][2] = "nn.Identity"

        # Define model
        ch = self.yaml["ch"] = self.yaml.get("ch", ch)  # input channels
        if nc and nc != self.yaml["nc"]:
            self.yaml["nc"] = nc  # override YAML value
        self.model, self.save = parse_model(deepcopy(self.yaml), ch=ch, verbose=verbose)  # model, savelist
        self.names = {i: f"{i}" for i in range(self.yaml["nc"])}  # default names dict
        self.inplace = self.yaml.get("inplace", True)
        self.end2end = getattr(self.model[-1], "end2end", False)

        # Build strides
        m = self.model[-1]  # Detect()
        if isinstance(m, Detect):  # includes all Detect subclasses like Segment, Pose, OBB, YOLOEDetect, YOLOESegment
            s = 256  # 2x min stride
            m.inplace = self.inplace

            def _forward(x):
                """Performs a forward pass through the model, handling different Detect subclass types accordingly."""
                if self.end2end:
                    return self.forward(x)["one2many"]
                return self.forward(x)[0] if isinstance(m, (Segment, YOLOESegment, Pose, OBB)) else self.forward(x)

            m.stride = torch.tensor([s / x.shape[-2] for x in _forward(torch.zeros(1, ch, s, s))])  # forward
            self.stride = m.stride
            m.bias_init()  # only run once
        else:
            self.stride = torch.Tensor([32])  # default stride for i.e. RTDETR

        # Init weights, biases
        initialize_weights(self)
        if verbose:
            self.info()

    def _predict_augment(self, x):
        """Perform augmentations on input image x and return augmented inference and train outputs."""
        if getattr(self, "end2end", False) or self.__class__.__name__ != "DetectionModel":
            return self._predict_once(x)
        img_size = x.shape[-2:]  # height, width
        s = [1, 0.83, 0.67]  # scales
        f = [None, 3, None]  # flips (2-ud, 3-lr)
        y = []  # outputs
        for si, fi in zip(s, f):
            xi = scale_img(x.flip(fi) if fi else x, si, gs=int(self.stride.max()))
            yi = super().predict(xi)[0]  # forward
            yi = self._descale_pred(yi, fi, si, img_size)
            y.append(yi)
        y = self._clip_augmented(y)  # clip augmented tails
        return torch.cat(y, -1), None  # augmented inference, train

    @staticmethod
    def _descale_pred(p, flips, scale, img_size, dim=1):
        """De-scale predictions following augmented inference (inverse operation)."""
        p[:, :4] /= scale  # de-scale
        x, y, wh, cls = p.split((1, 1, 2, p.shape[dim] - 4), dim)
        if flips == 2:
            y = img_size[0] - y  # de-flip ud
        elif flips == 3:
            x = img_size[1] - x  # de-flip lr
        return torch.cat((x, y, wh, cls), dim)

    def _clip_augmented(self, y):
        """Clip YOLO augmented inference tails."""
        nl = self.model[-1].nl  # number of detection layers (P3-P5)
        g = sum(4**x for x in range(nl))  # grid points
        e = 1  # exclude layer count
        i = (y[0].shape[-1] // g) * sum(4**x for x in range(e))  # indices
        y[0] = y[0][..., :-i]  # large
        i = (y[-1].shape[-1] // g) * sum(4 ** (nl - 1 - x) for x in range(e))  # indices
        y[-1] = y[-1][..., i:]  # small
        return y

    def init_criterion(self):
        """Initialize the loss criterion for the DetectionModel."""
        return E2EDetectLoss(self) if getattr(self, "end2end", False) else v8DetectionLoss(self)


class SegmentationModel(DetectionModel):
    """YOLOv8 segmentation model."""

    def __init__(self, cfg="yolov8n-seg.yaml", ch=3, nc=None, verbose=True):
        """Initialize YOLOv8 segmentation model with given config and parameters."""
        super().__init__(cfg=cfg, ch=ch, nc=nc, verbose=verbose)

    def init_criterion(self):
        """Initialize the loss criterion for the SegmentationModel."""
        return v8SegmentationLoss(self)


class YOLOEModel(DetectionModel):
    """YOLOE Model."""

    def __init__(self, cfg="yoloe-v8s.yaml", ch=3, nc=None, verbose=True):
        """Initialize YOLOE model with given config and parameters."""
        # Randomness
        # self.txt_feats = torch.randn(1, nc or 80, 512)  # features placeholder
        self.clip_model = None  # CLIP model placeholder
        super().__init__(cfg=cfg, ch=ch, nc=nc, verbose=verbose)
    
    @smart_inference_mode()
    def get_text_pe(self, text, batch=80, cache_clip_model=False):
        assert(not self.training)
        
        """Set classes in advance so that model could do offline-inference without clip model."""
        from visaionlibrary.utils.yoloe.nn.text_model import build_text_model
        
        device = next(self.model.parameters()).device
        
        text_model = self.args.get("text_model")
        if (
            not getattr(self, "clip_model", None) and cache_clip_model
        ):  # for backwards compatibility of models lacking clip_model attribute
            self.clip_model = build_text_model(text_model, device=device)
            
        model = self.clip_model if cache_clip_model else build_text_model(text_model, device=device)
        text_token = model.tokenize(text)
        txt_feats = model.encode_text(text_token)
        txt_feats = txt_feats.reshape(-1, len(text), txt_feats.shape[-1])
        
        head = self.model[-1]
        assert(isinstance(head, YOLOEDetect))
        return head.get_tpe(txt_feats)
    
    @smart_inference_mode()
    def get_visual_pe(self, img, visual):
        return self(img, vpe=visual, return_vpe=True)

    def set_vocab(self, vocab, names):
        assert(not self.training)
        head = self.model[-1]
        assert(isinstance(head, YOLOEDetect))
        
        # Cache anchors for head
        device = next(self.parameters()).device
        self(torch.empty(1, 3, self.args["imgsz"], self.args["imgsz"]).to(device))  # warmup
        
        self.model[-1].lrpc = nn.ModuleList(LRPCHead(cls, pf[-1], loc[-1], enabled=i!=2) for i, (cls, pf, loc) in enumerate(zip(vocab, head.cv3, head.cv2)))
        for loc_head, cls_head in zip(head.cv2, head.cv3):
            assert(isinstance(loc_head, nn.Sequential))
            assert(isinstance(cls_head, nn.Sequential))
            del loc_head[-1]
            del cls_head[-1]
        self.model[-1].nc = len(names)
        self.names = check_class_names(names)

    def get_vocab(self, names):
        assert(not self.training)
        head = self.model[-1]
        assert(isinstance(head, YOLOEDetect))
        assert(not head.is_fused)
        
        tpe = self.get_text_pe(names)
        self.set_classes(names, tpe)
        self.fuse()

        vocab = nn.ModuleList()
        for cls_head in head.cv3:
            assert(isinstance(cls_head, nn.Sequential))
            vocab.append(cls_head[-1])
        return vocab

    def set_classes(self, names, embeddings):
        """Set classes in advance so that model could do offline-inference without clip model."""
        assert(embeddings.ndim == 3)
        self.pe = embeddings
        self.model[-1].nc = len(names)
        self.names = check_class_names(names)

    def get_cls_pe(self, tpe, vpe):
        all_pe = []
        if tpe is not None:
            assert(tpe.ndim == 3)
            all_pe.append(tpe)
        if vpe is not None:
            assert(vpe.ndim == 3)
            all_pe.append(vpe)
        if len(all_pe) == 0:
            all_pe.append(getattr(self, 'pe', torch.zeros(1, 80, 512)))
        return torch.cat(all_pe, dim=1)

    def predict(self, x, profile=False, visualize=False, tpe=None, \
        augment=False, embed=None, vpe=None, return_vpe=False):
        """
        Perform a forward pass through the model.

        Args:
            x (torch.Tensor): The input tensor.
            profile (bool, optional): If True, profile the computation time for each layer. Defaults to False.
            visualize (bool, optional): If True, save feature maps for visualization. Defaults to False.
            txt_feats (torch.Tensor): The text features, use it if it's given. Defaults to None.
            augment (bool, optional): If True, perform data augmentation during inference. Defaults to False.
            embed (list, optional): A list of feature vectors/embeddings to return.

        Returns:
            (torch.Tensor): Model's output tensor.
        """            
        y, dt, embeddings = [], [], []  # outputs
        b = x.shape[0]
        for m in self.model:  # except the head part
            if m.f != -1:  # if not from previous layer
                x = y[m.f] if isinstance(m.f, int) else [x if j == -1 else y[j] for j in m.f]  # from earlier layers
            if profile:
                self._profile_one_layer(m, x, dt)
            if isinstance(m, C2fAttn):
                x = m(x, tpe or getattr(self, 'pe', torch.zeros(1, 80, 512)).to(x.device))
            elif isinstance(m, YOLOEDetect):
                vpe = m.get_vpe(x, vpe) if vpe is not None else None
                if return_vpe:
                    assert(vpe is not None)
                    assert(not self.training)
                    return vpe
                cls_pe = self.get_cls_pe(m.get_tpe(tpe), vpe).to(device=x[0].device, dtype=x[0].dtype)
                if len(cls_pe) != b:
                    cls_pe = cls_pe.repeat(b, 1, 1)
                x = m(x, cls_pe)
            else:
                x = m(x)  # run

            y.append(x if m.i in self.save else None)  # save output

            if embed and m.i in embed:
                embeddings.append(nn.functional.adaptive_avg_pool2d(x, (1, 1)).squeeze(-1).squeeze(-1))  # flatten
                if m.i == max(embed):
                    return torch.unbind(torch.cat(embeddings, 1), dim=0)
        return x

    def loss(self, batch, preds=None):
        """
        Compute loss.

        Args:
            batch (dict): Batch to compute loss on.
            preds (torch.Tensor | List[torch.Tensor]): Predictions.
        """
        if not hasattr(self, "criterion"):
            self.criterion = self.init_criterion()

        if preds is None:
            preds = self.forward(batch["img"], tpe=batch.get("txt_feats", None), 
                                 vpe=batch["visuals"] if self.args.load_vp else None)
        return self.criterion(preds, batch)
    
    def init_criterion(self):
        if self.args.load_vp:
            return TVPDetectLoss(self)
        else:
            return super().init_criterion()


class YOLOESegModel(YOLOEModel, SegmentationModel):
    """YOLOE segmentation model."""

    def __init__(self, cfg="yoloe-v8s-seg.yaml", ch=3, nc=None, verbose=True):
        """Initialize YOLOE segmentation model with given config and parameters."""
        super().__init__(cfg=cfg, ch=ch, nc=nc, verbose=verbose)

    def init_criterion(self):
        """Initialize the loss criterion for the SegmentationModel."""
        if self.args.load_vp:
            return TVPSegmentLoss(self)
        else:
            return super().init_criterion()


class Ensemble(nn.ModuleList):
    """Ensemble of models."""

    def __init__(self):
        """Initialize an ensemble of models."""
        super().__init__()

    def forward(self, x, augment=False, profile=False, visualize=False):
        """Function generates the YOLO network's final layer."""
        y = [module(x, augment, profile, visualize)[0] for module in self]
        # y = torch.stack(y).max(0)[0]  # max ensemble
        # y = torch.stack(y).mean(0)  # mean ensemble
        y = torch.cat(y, 2)  # nms ensemble, y shape(B, HW, C)
        return y, None  # inference, train output


# Functions ------------------------------------------------------------------------------------------------------------


@contextlib.contextmanager
def temporary_modules(modules=None, attributes=None):
    """
    Context manager for temporarily adding or modifying modules in Python's module cache (`sys.modules`).

    This function can be used to change the module paths during runtime. It's useful when refactoring code,
    where you've moved a module from one location to another, but you still want to support the old import
    paths for backwards compatibility.

    Args:
        modules (dict, optional): A dictionary mapping old module paths to new module paths.
        attributes (dict, optional): A dictionary mapping old module attributes to new module attributes.

    Example:
        ```python
        with temporary_modules({"old.module": "new.module"}, {"old.module.attribute": "new.module.attribute"}):
            import old.module  # this will now import new.module
            from old.module import attribute  # this will now import new.module.attribute
        ```

    Note:
        The changes are only in effect inside the context manager and are undone once the context manager exits.
        Be aware that directly manipulating `sys.modules` can lead to unpredictable results, especially in larger
        applications or libraries. Use this function with caution.
    """
    if modules is None:
        modules = {}
    if attributes is None:
        attributes = {}
    import sys
    from importlib import import_module

    try:
        # Set modules in sys.modules under their old name FIRST
        # This must happen before setting attributes, as attributes may reference these modules
        for old, new in modules.items():
            sys.modules[old] = import_module(new)

        # Set attributes in sys.modules under their old name
        for old, new in attributes.items():
            old_module, old_attr = old.rsplit(".", 1)
            new_module, new_attr = new.rsplit(".", 1)
            setattr(import_module(old_module), old_attr, getattr(import_module(new_module), new_attr))

        yield
    finally:
        # Remove the temporary module paths
        for old in modules:
            if old in sys.modules:
                del sys.modules[old]


class SafeClass:
    """A placeholder class to replace unknown classes during unpickling."""

    def __init__(self, *args, **kwargs):
        """Initialize SafeClass instance, ignoring all arguments."""
        pass

    def __call__(self, *args, **kwargs):
        """Run SafeClass instance, ignoring all arguments."""
        pass


class SafeUnpickler(pickle.Unpickler):
    """Custom Unpickler that replaces unknown classes with SafeClass."""

    def find_class(self, module, name):
        """Attempt to find a class, returning SafeClass if not among safe modules."""
        safe_modules = (
            "torch",
            "collections",
            "collections.abc",
            "builtins",
            "math",
            "numpy",
            # Add other modules considered safe
        )
        if module in safe_modules:
            return super().find_class(module, name)
        else:
            return SafeClass


def torch_safe_load(weight, safe_only=False):
    """
    Attempts to load a PyTorch model with the torch.load() function. If a ModuleNotFoundError is raised, it catches the
    error, logs a warning message, and attempts to install the missing module via the () function.
    After installation, the function again attempts to load the model using torch.load().

    Args:
        weight (str): The file path of the PyTorch model.
        safe_only (bool): If True, replace unknown classes with SafeClass during loading.

    Returns:
        ckpt (dict): The loaded model checkpoint.
        file (str): The loaded filename
    """
    try:
        with temporary_modules(
            modules={
                "ultralytics": "visaionlibrary.utils.yoloe",
            },
            attributes={
            },
        ):
            if safe_only:
                # Load via custom pickle module
                safe_pickle = types.ModuleType("safe_pickle")
                safe_pickle.Unpickler = SafeUnpickler
                safe_pickle.load = lambda file_obj: SafeUnpickler(file_obj).load()
                with open(weight, "rb") as f:
                    ckpt = torch.load(f, pickle_module=safe_pickle)
            else:
                ckpt = torch.load(weight, map_location="cpu")

    except ModuleNotFoundError as e:  # e.name is missing module name
        raise e

    return ckpt, weight

def attempt_load_weights(weights, device=None, inplace=True, fuse=False):
    """Loads an ensemble of models weights=[a,b,c] or a single model weights=[a] or weights=a."""
    ensemble = Ensemble()
    for w in weights if isinstance(weights, list) else [weights]:
        ckpt, w = torch_safe_load(w)  # load ckpt
        
        args = {**DEFAULT_CFG_DICT, **ckpt["train_args"]} if "train_args" in ckpt else None  # combined args
        model = (ckpt.get("ema") or ckpt["model"]).to(device).float()  # FP32 model

        # Model compatibility updates
        model.args = args  # attach args to model
        model.pt_path = w  # attach *.pt file path to model
        model.task = guess_model_task(model)
        if not hasattr(model, "stride"):
            model.stride = torch.tensor([32.0])

        # Append
        ensemble.append(model.fuse().eval() if fuse and hasattr(model, "fuse") else model.eval())  # model in eval mode

    # Module updates
    for m in ensemble.modules():
        if hasattr(m, "inplace"):
            m.inplace = inplace
        elif isinstance(m, nn.Upsample) and not hasattr(m, "recompute_scale_factor"):
            m.recompute_scale_factor = None  # torch 1.11.0 compatibility

    # Return model
    if len(ensemble) == 1:
        return ensemble[-1]

    # Return ensemble
    for k in "names", "nc", "yaml":
        setattr(ensemble, k, getattr(ensemble[0], k))
    ensemble.stride = ensemble[int(torch.argmax(torch.tensor([m.stride.max() for m in ensemble])))].stride
    assert all(ensemble[0].nc == m.nc for m in ensemble), f"Models differ in class counts {[m.nc for m in ensemble]}"
    return ensemble


def attempt_load_one_weight(weight, device=None, inplace=True, fuse=False):
    """Loads a single model weights."""
    ckpt, weight = torch_safe_load(weight)  # load ckpt
    args = {**DEFAULT_CFG_DICT, **(ckpt.get("train_args", {}))}  # combine model and default args, preferring model args
    model = (ckpt.get("ema") or ckpt["model"]).to(device).float()  # FP32 model

    # Model compatibility updates
    model.args = {k: v for k, v in args.items() if k in DEFAULT_CFG_KEYS}  # attach args to model
    model.pt_path = weight  # attach *.pt file path to model
    model.task = guess_model_task(model)
    if not hasattr(model, "stride"):
        model.stride = torch.tensor([32.0])

    model = model.fuse().eval() if fuse and hasattr(model, "fuse") else model.eval()  # model in eval mode

    # Module updates
    for m in model.modules():
        if hasattr(m, "inplace"):
            m.inplace = inplace
        elif isinstance(m, nn.Upsample) and not hasattr(m, "recompute_scale_factor"):
            m.recompute_scale_factor = None  # torch 1.11.0 compatibility

    # Return model and ckpt
    return model, ckpt


def parse_model(d, ch, verbose=True):  # model_dict, input_channels(3)
    """Parse a YOLO model.yaml dictionary into a PyTorch model."""
    import ast

    # Args
    legacy = True  # backward compatibility for v3/v5/v8/v9 models
    max_channels = float("inf")
    nc, act, scales = (d.get(x) for x in ("nc", "activation", "scales"))
    depth, width, kpt_shape = (d.get(x, 1.0) for x in ("depth_multiple", "width_multiple", "kpt_shape"))
    if scales:
        scale = d.get("scale")
        if not scale:
            scale = tuple(scales.keys())[0]
        depth, width, max_channels = scales[scale]

    if act:
        Conv.default_act = eval(act)  # redefine default activation, i.e. Conv.default_act = nn.SiLU()

    ch = [ch]
    layers, save, c2 = [], [], ch[-1]  # layers, savelist, ch out
    for i, (f, n, m, args) in enumerate(d["backbone"] + d["head"]):  # from, number, module, args
        m = getattr(torch.nn, m[3:]) if "nn." in m else globals()[m]  # get module
        for j, a in enumerate(args):
            if isinstance(a, str):
                with contextlib.suppress(ValueError):
                    args[j] = locals()[a] if a in locals() else ast.literal_eval(a)
        n = n_ = max(round(n * depth), 1) if n > 1 else n  # depth gain
        if m in {
            Classify,
            Conv,
            ConvTranspose,
            GhostConv,
            Bottleneck,
            GhostBottleneck,
            SPP,
            SPPF,
            C2fPSA,
            C2PSA,
            DWConv,
            Focus,
            BottleneckCSP,
            C1,
            C2,
            C2f,
            C3k2,
            RepNCSPELAN4,
            ELAN1,
            ADown,
            AConv,
            SPPELAN,
            C2fAttn,
            C3,
            C3TR,
            C3Ghost,
            nn.ConvTranspose2d,
            DWConvTranspose2d,
            C3x,
            RepC3,
            PSA,
            SCDown,
            C2fCIB,
        }:
            c1, c2 = ch[f], args[0]
            if c2 != nc:  # if c2 not equal to number of classes (i.e. for Classify() output)
                c2 = make_divisible(min(c2, max_channels) * width, 8)
            if m is C2fAttn:
                args[1] = make_divisible(min(args[1], max_channels // 2) * width, 8)  # embed channels
                args[2] = int(
                    max(round(min(args[2], max_channels // 2 // 32)) * width, 1) if args[2] > 1 else args[2]
                )  # num heads

            args = [c1, c2, *args[1:]]
            if m in {
                BottleneckCSP,
                C1,
                C2,
                C2f,
                C3k2,
                C2fAttn,
                C3,
                C3TR,
                C3Ghost,
                C3x,
                RepC3,
                C2fPSA,
                C2fCIB,
                C2PSA,
            }:
                args.insert(2, n)  # number of repeats
                n = 1
            if m is C3k2:  # for M/L/X sizes
                legacy = False
                if scale in "mlx":
                    args[3] = True
            if m is C2fCIB:
                legacy = False
        elif m is AIFI:
            args = [ch[f], *args]
        elif m in {HGStem, HGBlock}:
            c1, cm, c2 = ch[f], args[0], args[1]
            args = [c1, cm, c2, *args[2:]]
            if m is HGBlock:
                args.insert(4, n)  # number of repeats
                n = 1
        elif m is ResNetLayer:
            c2 = args[1] if args[3] else args[1] * 4
        elif m is nn.BatchNorm2d:
            args = [ch[f]]
        elif m is Concat:
            c2 = sum(ch[x] for x in f)
        elif m in {Detect, YOLOEDetect, Segment, YOLOESegment, Pose, OBB, ImagePoolingAttn, v10Detect}:
            args.append([ch[x] for x in f])
            if m is Segment or m is YOLOESegment:
                args[2] = make_divisible(min(args[2], max_channels) * width, 8)
            if m in {Detect, YOLOEDetect, Segment, YOLOESegment, Pose, OBB}:
                m.legacy = legacy
        elif m is RTDETRDecoder:  # special case, channels arg must be passed in index 1
            args.insert(1, [ch[x] for x in f])
        elif m is CBLinear:
            c2 = args[0]
            c1 = ch[f]
            args = [c1, c2, *args[1:]]
        elif m is CBFuse:
            c2 = ch[f[-1]]
        else:
            c2 = ch[f]

        m_ = nn.Sequential(*(m(*args) for _ in range(n))) if n > 1 else m(*args)  # module
        t = str(m)[8:-2].replace("__main__.", "")  # module type
        m_.np = sum(x.numel() for x in m_.parameters())  # number params
        m_.i, m_.f, m_.type = i, f, t  # attach index, 'from' index, type
        save.extend(x % i for x in ([f] if isinstance(f, int) else f) if x != -1)  # append to savelist
        layers.append(m_)
        if i == 0:
            ch = []
        ch.append(c2)
    return nn.Sequential(*layers), sorted(save)


def yaml_model_load(path):
    """Load a YOLOv8 model from a YAML file."""
    path = Path(path)
    if path.stem in (f"yolov{d}{x}6" for x in "nsmlx" for d in (5, 8)):
        new_stem = re.sub(r"(\d+)([nslmx])6(.+)?$", r"\1\2-p6\3", path.stem)
        path = path.with_name(new_stem + path.suffix)

    unified_path = re.sub(r"(\d+)([nslmx])(.+)?$", r"\1\3", str(path))  # i.e. yolov8x.yaml -> yolov8.yaml
    yaml_file = check_yaml(unified_path, hard=False) or check_yaml(path)
    d = yaml_load(yaml_file)  # model dict
    d["scale"] = guess_model_scale(path)
    d["yaml_file"] = str(path)
    return d


def guess_model_scale(model_path):
    """
    Takes a path to a YOLO model's YAML file as input and extracts the size character of the model's scale. The function
    uses regular expression matching to find the pattern of the model scale in the YAML file name, which is denoted by
    n, s, m, l, or x. The function returns the size character of the model scale as a string.

    Args:
        model_path (str | Path): The path to the YOLO model's YAML file.

    Returns:
        (str): The size character of the model's scale, which can be n, s, m, l, or x.
    """
    try:
        return re.search(r"yolo(e-)?[v]?\d+([nslmx])", Path(model_path).stem).group(2)  # noqa, returns n, s, m, l, or x
    except AttributeError:
        return ""


def guess_model_task(model):
    """
    Guess the task of a PyTorch model from its architecture or configuration.

    Args:
        model (nn.Module | dict): PyTorch model or model configuration in YAML format.

    Returns:
        (str): Task of the model ('detect', 'segment', 'classify', 'pose').

    Raises:
        SyntaxError: If the task of the model could not be determined.
    """

    def cfg2task(cfg):
        """Guess from YAML dictionary."""
        m = cfg["head"][-1][-2].lower()  # output module name
        if m in {"classify", "classifier", "cls", "fc"}:
            return "classify"
        if "detect" in m:
            return "detect"
        if "segment" in m:
            return "segment"
        if m == "pose":
            return "pose"
        if m == "obb":
            return "obb"

    # Guess from model cfg
    if isinstance(model, dict):
        with contextlib.suppress(Exception):
            return cfg2task(model)
    # Guess from PyTorch model
    if isinstance(model, nn.Module):  # PyTorch model
        for x in "model.args", "model.model.args", "model.model.model.args":
            with contextlib.suppress(Exception):
                return eval(x)["task"]
        for x in "model.yaml", "model.model.yaml", "model.model.model.yaml":
            with contextlib.suppress(Exception):
                return cfg2task(eval(x))
        for m in model.modules():
            if isinstance(m, (Segment, YOLOESegment)):
                return "segment"
            elif isinstance(m, Classify):
                return "classify"
            elif isinstance(m, Pose):
                return "pose"
            elif isinstance(m, OBB):
                return "obb"
            elif isinstance(m, (Detect, YOLOEDetect, v10Detect)):
                return "detect"

    # Guess from model filename
    if isinstance(model, (str, Path)):
        model = Path(model)
        if "-seg" in model.stem or "segment" in model.parts:
            return "segment"
        elif "-cls" in model.stem or "classify" in model.parts:
            return "classify"
        elif "-pose" in model.stem or "pose" in model.parts:
            return "pose"
        elif "-obb" in model.stem or "obb" in model.parts:
            return "obb"
        elif "detect" in model.parts:
            return "detect"

    return "detect"  # assume detect



from visaionlibrary.datasets.utils.sample_status import SampleStatus
from visaionlibrary.datasets.utils.str2mask import decode_str2mask, encode_mask_to_str

__all__ = ["SampleStatus", "decode_str2mask", "encode_mask_to_str"]


__version__ = "0.6.30"

from visaionlibrary.datasets import *
from visaionlibrary.utils import *
from visaionlibrary.evaluation import *
from visaionlibrary.models import *
from visaionlibrary.engine import *



import math
from typing import List, Tuple, Union, Sequence
import torch
import torch.nn as nn
from torch import Tensor
from mmengine.model import BaseModule
from mmdet.utils import OptMultiConfig
from mmdet.models.utils import multi_apply
from mmyolo.models.utils import make_divisible
from mmengine.registry import MODELS
from ..layers.yolo_bricks import Proto, Conv, DWConv

@MODELS.register_module()
class YOLO11InsHeadModule(BaseModule):
    def __init__(self,
                 num_classes: int,
                 in_channels: Union[int, Sequence],
                 widen_factor: float = 1.0,
                 num_base_priors: int = 1,
                 mask_channels: int = 32,
                 proto_channels: int = 256,
                 featmap_strides: Sequence[int] = (8, 16, 32),
                 init_cfg: OptMultiConfig = None):
        super().__init__(init_cfg=init_cfg)
        self.num_classes = num_classes
        self.widen_factor = widen_factor
        self.mask_channels = mask_channels
        # self.num_out_attrib_with_proto = 5 + num_classes + mask_channels
        self.proto_channels = make_divisible(proto_channels, widen_factor)
        self.featmap_strides = featmap_strides
        self.num_out_attrib = 5 + self.num_classes
        self.num_levels = len(self.featmap_strides)
        self.num_base_priors = num_base_priors

        if isinstance(in_channels, int):
            self.in_channels = [make_divisible(in_channels, widen_factor)
                                ] * self.num_levels
        else:
            self.in_channels = [
                make_divisible(i, widen_factor) for i in in_channels
            ]

        self._init_layers()


    def _init_layers(self):
        """initialize conv layers in YOLOv5 Ins head."""
        reg_out_channels = max(16, self.in_channels[0] // 4)
        cls_out_channels = max(self.in_channels[0], self.num_classes)
        c4 = max(self.in_channels[0] // 4, self.num_classes)
        self.cv2 = nn.ModuleList()
        self.cv3 = nn.ModuleList()
        self.cv4 = nn.ModuleList()
        self.proto = Proto(
            self.in_channels[0],
            self.proto_channels,
            self.mask_channels)
        for i in range(self.num_levels):
            self.cv2.append(
                nn.Sequential(
                    Conv(self.in_channels[i], reg_out_channels, 3),
                    Conv(reg_out_channels, reg_out_channels, 3),
                    nn.Conv2d(reg_out_channels, 4 * self.num_base_priors, 1))
            )  # box
            self.cv3.append(
                    nn.Sequential(
                        nn.Sequential(DWConv(self.in_channels[i], self.in_channels[i], 3), Conv(self.in_channels[i], cls_out_channels, 1)),
                        nn.Sequential(DWConv(cls_out_channels, cls_out_channels, 3), Conv(cls_out_channels, cls_out_channels, 1)),
                        nn.Conv2d(cls_out_channels, self.num_classes * self.num_base_priors, 1),
                    )
                )  # cls
            self.cv4.append(nn.Sequential(Conv(self.in_channels[i], c4, 3), Conv(c4, c4, 3), nn.Conv2d(c4, self.mask_channels * self.num_base_priors, 1)))

    def init_weights(self):
        """Initialize the bias of YOLOv5 head."""
        super().init_weights()
        for c2, c3, c4, s in zip(self.cv2, self.cv3, self.cv4, self.featmap_strides):  # from

            c2[-1].bias.data[:] = 1.0 # box
            c3[-1].bias.data[:self.num_classes] = math.log(
                5 / self.num_classes / (640 / s)**2)
            c4[-1].bias.data[:] = 1.0 # mask


    def forward(self, x: Tuple[Tensor]) -> Tuple[List]:
        """Forward features from the upstream network.

        Args:
            x (Tuple[Tensor]): Features from the upstream network, each is
                a 4D-tensor.
        Returns:
            Tuple[List]: A tuple of multi-level classification scores, bbox
            predictions, objectnesses, and mask predictions.
        """
        mask_protos = self.proto(x[0])  # mask protos (b, n, 160, 160)
        cls_scores, bbox_preds, objectnesses, coeff_preds = multi_apply(
            self.forward_single, x, self.cv2, self.cv3, self.cv4)
        
        return cls_scores, bbox_preds, objectnesses, coeff_preds, mask_protos

    def forward_single(
            self, x: Tensor,
            cv2: nn.Module, cv3: nn.Module, cv4: nn.Module) -> Tuple[Tensor, Tensor, Tensor, Tensor]:

        """Forward feature of a single scale level."""

        bs, _, ny, nx = x.shape
        cls_score = cv3(x).view(bs, -1, ny, nx)  # (b, 3*c, ny, nx)
        c = cls_score.shape[1] // self.num_base_priors
        bbox_pred = cv2(x).view(bs, -1, ny, nx)  # (b, 12, ny, nx)
        objectness = [cls_score[:, i*c: (i+1)*c].max(dim=1, keepdim=True).values for i in range(self.num_base_priors)]
        objectness = torch.concat(objectness, dim=1) # (b, 3, ny, nx)
        coeff_pred = cv4(x).view(bs, -1, ny, nx)  # (b, _, ny, nx)

        return cls_score, bbox_pred, objectness, coeff_pred


from visaionlibrary.datasets.det_dataset import VisaionDetDataset
from visaionlibrary.datasets.cls_dataset import VisaionClsDataset
from visaionlibrary.datasets.seg_dataset import VisaionSegDataset
from visaionlibrary.datasets.ins_dataset import VisaionYOLOv5InsDataset, VisaionRTMDETInsDataset
from visaionlibrary.datasets.samplers import VisaionInfiniteSampler, InfiniteSamplerBatch, BatchSamplerConstantOne
from visaionlibrary.datasets.transforms import LoadSegAnnotations, RandomRescale, SegRegionSampler, PackSegInputsVisaion

__all__ = [
    "VisaionDetDataset", "VisaionClsDataset", "VisaionSegDataset", "VisaionYOLOv5InsDataset", "VisaionRTMDETInsDataset",
    "VisaionInfiniteSampler", "InfiniteSamplerBatch", "BatchSamplerConstantOne", 
    "LoadSegAnnotations", "RandomRescale", "SegRegionSampler", "PackSegInputsVisaion"
]


from typing import List, Union

import torch.nn as nn
from mmdet.utils import ConfigType, OptMultiConfig
from ..layers.yolo_bricks import C3K2, Conv, Concat
from mmyolo.models.necks import YOLOv8PAFPN
from mmyolo.models.utils import make_divisible, make_round
from mmengine.registry import MODELS

@MODELS.register_module()
class YOLO11PAFPN(YOLOv8PAFPN):
    """YOLO11 Path Aggregation Feature Pyramid Network (PAFPN) neck.

    相比 YOLOv8PAFPN，YOLO11PAFPN 主要做了以下改动：
    1. 将 C2f 模块替换为 C3K2，提升梯度流与特征复用效率；
    2. 引入可选的轻量级注意力机制（如 SE、CBAM），增强多尺度特征表达能力；
    3. 支持更灵活的通道缩放因子，便于模型瘦身与部署；
    4. 优化上/下采样路径，减少冗余计算，提升小目标检测性能。
    """

    def __init__(self,
                 in_channels: List[int],
                 out_channels: int,
                 deepen_factor: float = 1.0,
                 widen_factor: float = 1.0,
                 num_csp_blocks: int = 2,
                 use_attention: str = None,          # 'SE', 'CBAM', None
                 freeze_all: bool = False,
                 norm_cfg: ConfigType = dict(type='BN', momentum=0.03, eps=0.001),
                 act_cfg: ConfigType = dict(type='SiLU', inplace=True),
                 init_cfg: OptMultiConfig = None) -> None:
        if widen_factor > 0.5: # foe xlm model size
            self.c3k = True
        else:
            self.c3k = False
        # 用 C3K2 替换基类中的 C2f
        super().__init__(
            in_channels=in_channels,
            out_channels=out_channels,
            deepen_factor=deepen_factor,
            widen_factor=widen_factor,
            num_csp_blocks=num_csp_blocks,
            freeze_all=freeze_all,
            norm_cfg=norm_cfg,
            act_cfg=act_cfg,
            init_cfg=init_cfg)
        

    def build_reduce_layer(self, idx: int) -> nn.Module:
        """build reduce layer.

        Args:
            idx (int): layer idx.

        Returns:
            nn.Module: The reduce layer.
        """
        # if idx == len(self.in_channels) - 1:
        #     layer = Conv(
        #         make_divisible(self.in_channels[idx], self.widen_factor),
        #         make_divisible(self.in_channels[idx - 1], self.widen_factor),
        #         1)
        # else:
        #     layer = nn.Identity()
        layer = nn.Identity()  # yolo11 中没有 reduce layer
        return layer

    def build_downsample_layer(self, idx: int) -> nn.Module:
        """build downsample layer.
        layer 0: 256,256
        layer 1: 512,512

        Args:
            idx (int): layer idx.

        Returns:
            nn.Module: The downsample layer.
        """
        return Conv(
            make_divisible(self.out_channels[idx], self.widen_factor),
            make_divisible(self.out_channels[idx], self.widen_factor),
            k=3,
            s=2)

    def build_top_down_layer(self, idx: int) -> nn.Module:
        """build top down layer.
        layer 1: 512+512 ,256
        layer 2:  512+1024 ,512
        Args:
            idx (int): layer idx.

        Returns:
            nn.Module: The top down layer.
        """
        return C3K2(
            make_divisible((self.in_channels[idx - 1] + self.in_channels[idx]),
                           self.widen_factor),
            make_divisible(self.out_channels[idx - 1], self.widen_factor),
            num_blocks=make_round(self.num_csp_blocks, self.deepen_factor),
            c3k=self.c3k)

    def build_bottom_up_layer(self, idx: int) -> nn.Module:
        """build bottom up layer.
        layer 0:512+256,512
        layer 1:last+512,1024
        Args:
            idx (int): layer idx.

        Returns:
            nn.Module: The bottom up layer.
        """
        if idx == 1:
            c3k = True
        else:
            c3k = self.c3k
        return C3K2(
            make_divisible(
                (self.out_channels[idx] + self.out_channels[idx + 1]),
                self.widen_factor),
            make_divisible(self.out_channels[idx + 1], self.widen_factor),
            num_blocks=make_round(self.num_csp_blocks, self.deepen_factor),
            c3k=c3k)
